
from ceramatch.view.vcontrols_groups.controls.button import (Button)
from ceramatch.view.vcontrols_groups.controls.combo import (Combo)

from PySide2 import (QtWidgets, QtCore, QtGui)

class HeaderButton(QtWidgets.QToolButton):
	
	def __init__(self, label):
		
		QtWidgets.QToolButton.__init__(self)
		
		self.setText(label)
		self.setArrowType(QtCore.Qt.RightArrow)
		self.setSizePolicy(
			QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed
		)
		self.setAutoRaise(True)
		self.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
		self.setStyleSheet('''
			QToolButton {
				border-top: 1px solid white; 
				border-left: 1px solid white; border-bottom: 1px solid gray; 
				border-right: 1px solid gray;
			} 
			QToolButton:pressed {
				border-top: 1px solid gray; border-left: 1px solid gray; 
				border-bottom: 1px solid white; border-right: 1px solid white;
			}
		''')

class Descriptors(QtWidgets.QGroupBox):
	
	signal_load_drawings = QtCore.Signal()
	signal_cluster_classes_changed = QtCore.Signal()
	
	def __init__(self):
		
		QtWidgets.QGroupBox.__init__(self, "Classes / Descriptors")
		
		self.setStyleSheet("QGroupBox {font-weight: bold;}")
		self.setLayout(QtWidgets.QVBoxLayout())
		
		self.cluster_class_combo = Combo(
			self.on_cluster_class_changed, editable = True
		)
		self.node_class_combo = Combo(
			self.on_node_class_changed, editable = True
		)
		self.position_class_combo = Combo(
			self.on_position_class_changed, editable = True
		)
		
		classes_frame = QtWidgets.QFrame()
		classes_frame.setLayout(QtWidgets.QFormLayout())
		classes_frame.layout().setContentsMargins(0, 0, 0, 0)
		
		classes_frame.layout().addRow(
			QtWidgets.QLabel("Cluster Class:"), self.cluster_class_combo
		)
		classes_frame.layout().addRow(
			QtWidgets.QLabel("Node Class:"), self.node_class_combo
		)
		classes_frame.layout().addRow(
			QtWidgets.QLabel("Position Descriptor:"), self.position_class_combo
		)
		
		self.descriptors_frame = QtWidgets.QFrame()
		self.descriptors_frame.setLayout(QtWidgets.QVBoxLayout())
		self.descriptors_frame.layout().setContentsMargins(0, 0, 0, 0)
		
		self.descriptors_form = QtWidgets.QFrame()
		self.descriptors_form.setLayout(QtWidgets.QFormLayout())
		self.descriptors_form.layout().setContentsMargins(0, 0, 0, 0)
		
		self.descriptors_frame.layout().addWidget(self.descriptors_form)
		
		self.load_drawings_button = Button("Load Drawings", self.on_load_drawings)
		
		descriptor_box = QtWidgets.QFrame()
		descriptor_box.setLayout(QtWidgets.QVBoxLayout())
		descriptor_box.layout().setContentsMargins(0, 0, 0, 0)
		self.descriptor_header_button = HeaderButton("Descriptors")
		descriptor_box.layout().addWidget(self.descriptor_header_button)
		descriptor_box.layout().addWidget(self.descriptors_frame)
		self.descriptor_header_button.clicked.connect(
			self.on_descriptors_clicked
		)
		
		self.layout().addWidget(classes_frame)
		self.layout().addWidget(descriptor_box)
		self.layout().addWidget(self.load_drawings_button)
		
		self.descriptors_frame.setVisible(False)
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot()
	def on_descriptors_clicked(self):
		
		visible = self.descriptors_frame.isVisible()
		if visible:
			self.descriptor_header_button.setArrowType(QtCore.Qt.RightArrow)
		else:
			self.descriptor_header_button.setArrowType(QtCore.Qt.DownArrow)
		self.descriptors_frame.setVisible(not visible)
	
	@QtCore.Slot()
	def on_node_class_changed(self):
		
		self.signal_cluster_classes_changed.emit()
	
	@QtCore.Slot()
	def on_position_class_changed(self):
		
		self.signal_cluster_classes_changed.emit()
	
	@QtCore.Slot()
	def on_cluster_class_changed(self):
		
		self.signal_cluster_classes_changed.emit()
	
	@QtCore.Slot()
	def on_load_drawings(self):
		
		self.signal_load_drawings.emit()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def clear(self):
		
		self.cluster_class_combo.clear_values()
		self.node_class_combo.clear_values()
		self.position_class_combo.clear_values()
		self.set_descriptors({})
	
	
	def set_cluster_class(self, name, classes):
		
		self.cluster_class_combo.set_values(classes, name)
	
	def set_node_class(self, name, classes):
		
		self.node_class_combo.set_values(classes, name)
	
	def set_position_class(self, name, classes):
		
		self.position_class_combo.set_values(classes, name)
	
	def set_descriptors(self, descriptors):
		# descriptors = {name: chain, ...}
		
		for row in range(self.descriptors_form.layout().rowCount())[::-1]:
			self.descriptors_form.layout().removeRow(row)
		for name in descriptors:
			self.descriptors_form.layout().addRow(
				QtWidgets.QLabel("   %s" % (name)), 
				QtWidgets.QLabel(descriptors[name]),
			)
	
	def set_load_drawings_enabled(self, state):
		
		self.load_drawings_button.setEnabled(state)
	
	def get_cluster_class(self):
		
		return self.cluster_class_combo.get_value()
	
	def get_node_class(self):
		
		return self.node_class_combo.get_value()
	
	def get_position_class(self):
		
		return self.position_class_combo.get_value()




from ceramatch.dialogs.dialog_connect import DialogConnect
from ceramatch.dialogs.dialog_import_clustering import DialogImportClustering
from ceramatch.dialogs.dialog_export_drawing import DialogExportDrawing
from ceramatch.dialogs.dialog_about import DialogAbout

from ceramatch.utils.fnc_matching import (calc_distances)
from deposit.utils.fnc_files import (sanitize_filename)

from deposit_gui import DCDialogs

from PySide2 import (QtWidgets, QtCore, QtGui)
import os

class CDialogs(DCDialogs):
	
	def __init__(self, cmain, cview):
		
		DCDialogs.__init__(self, cmain, cview)
		
		self.view = cview._view
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot()
	def on_clear_recent(self):
		
		self.view.clear_recent_connections()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	
	
	# ---- Dialogs
	# ------------------------------------------------------------------------
	'''
	open(name, *args, **kwargs)
	
	Implement set_up_[name], process_[name] and cancel_[name] for each dialog:
	
	def set_up_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		dialog = QtWidgets.QDialog
		
		dialog.set_title(name)
		dialog.set_frame(frame = QtWidget)
		dialog.get_frame()
		dialog.set_button_box(ok: bool, cancel: bool): set if OK and Cancel buttons are visible
		
	def process_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		process dialog after OK has been clicked
	
	def cancel_[name](self, dialog, *args, **kwargs):
		
		args and kwargs are passed from DCDialogs.open(name, *args, **kwargs)
		
		handle dialog after cancel has been clicked
	'''
	
	def set_up_Connect(self, dialog, *args, **kwargs):
		
		frame = DialogConnect(dialog)
		frame.set_recent_dir(self.view.get_recent_dir())
		frame.set_recent_connections(self.view.get_recent_connections())
		frame.signal_clear_recent.connect(self.on_clear_recent)
	
	def process_Connect(self, dialog, *args, **kwargs):
		
		self.cmain.cmodel.load(**dialog.get_data())
	
	def cancel_Connect(self, dialog, *args, **kwargs):
		
		datasource = self.cmain.cmodel.get_datasource()
		if not datasource.is_valid():
			self.cmain.cmodel.load(datasource = "Memory")
	
	
	def set_up_SaveAsPostgres(self, dialog, *args, **kwargs):
		
		frame = DSaveAsPostgresFrame(dialog)
		frame.set_recent_connections(self.view.get_recent_connections())
	
	def process_SaveAsPostgres(self, dialog, *args, **kwargs):
		
		data = dialog.get_data()
		datasource = data["datasource"]
		self.cmain.cmodel.set_local_folder(datasource.get_local_folder())
		self.cmain.cmodel.save(**data)
		
		if self.cmain.cview.show_question(
			"Load Database?", 
			"Load database from <b>%s</b>" % (datasource.get_name()),
		):
			self.cmain.cmodel.load(**data)
	
	
	def set_up_SelectDistances(self, dialog):
		
		dialog.set_title("Select Distances to Calculate")
		dialog.setModal(True)
		dialog.set_button_box(True, True)
		
		frame = QtWidgets.QFrame()
		frame.setMinimumWidth(350)
		frame.setLayout(QtWidgets.QVBoxLayout())
		frame.check_diam = QtWidgets.QCheckBox("Diameter")
		frame.check_diam.setChecked(True)
		frame.check_ax = QtWidgets.QCheckBox("Axis")
		frame.check_ax.setChecked(True)
		frame.check_dice = QtWidgets.QCheckBox("Dice")
		frame.check_dice.setChecked(True)
		frame.check_dice_rim = QtWidgets.QCheckBox("Dice (rim only)")
		frame.check_dice_rim.setChecked(True)
		
		warning_frame = QtWidgets.QFrame()
		warning_frame.setLayout(QtWidgets.QHBoxLayout())
		icon = QtWidgets.QApplication.style().standardIcon(
			QtWidgets.QStyle.SP_MessageBoxWarning
		)
		label = QtWidgets.QLabel()
		label.setPixmap(icon.pixmap(QtCore.QSize(32, 32)))
		warning_frame.layout().addWidget(label)
		label = QtWidgets.QLabel(
			"Warning: Dice distance calculation is very computationally intensive"
		)
		label.setWordWrap(True)
		warning_frame.layout().addWidget(label)
		warning_frame.layout().addStretch()
		
		frame.layout().addWidget(frame.check_diam)
		frame.layout().addWidget(frame.check_ax)
		frame.layout().addWidget(frame.check_dice)
		frame.layout().addWidget(frame.check_dice_rim)
		frame.layout().addWidget(warning_frame)
		
		dialog.set_frame(frame)
	
	def process_SelectDistances(self, dialog):
		
		frame = dialog.get_frame()
		diam = frame.check_diam.isChecked()
		axis = frame.check_ax.isChecked()
		dice = frame.check_dice.isChecked()
		dice_rim = frame.check_dice_rim.isChecked()
		
		data = self.cmain.cgraph.get_profile_data()
		# data = {obj_id: (coords, radius), ...}
		
		select_components = [diam, axis, dice, dice_rim]
		calc_distances(
			data, select_components, self.cmain.cmodel._distance,
			progress = self.cmain.cview.progress,
		)
		
		self.cmain.cmodel.save_distance()
	
	
	def set_up_ImportClustering(self, dialog):
		
		dialog.set_title("")
		dialog.set_button_box(True, True)
		dialog.setModal(True)
		dialog.set_frame(DialogImportClustering(self.cmain.cview))
	
	def process_ImportClustering(self, dialog):
		
		frame = dialog.get_frame()
		path = frame.path_edit.text().strip()
		sample_column = frame.sample_combo.currentText()
		cluster_column = frame.cluster_combo.currentText()
		if (not sample_column) or (not cluster_column) or (not os.path.isfile(path)):
			return
		self.cmain.cgraph.import_clusters(path, sample_column, cluster_column)
	
	
	def set_up_ExportDendrogram(self, dialog):
		
		dialog.set_title("Export PDF Dendrogram")
		path = self.cmain.cview.get_recent_dir()
		if not path:
			path = self.cmain.cview.get_default_folder()
		filename = "%s.pdf" % sanitize_filename(
			self.cmain.cmodel.get_datasource_name() + "_dendrogram",
			"dendrogram",
		)
		frame = DialogExportDrawing(
			dialog,
			path = os.path.normpath(os.path.join(path, filename)),
			formats = "Adobe PDF (*.pdf)",
			multi = False,
			show_scale = False,
			show_dpi = True,
			show_page_size = True,
		)
		dialog.set_frame(frame)
	
	def process_ExportDendrogram(self, dialog):
		
		path, _, dpi, line_width, _, page_size = dialog.get_frame().get_data()
		if path is None:
			return
		
		self.cmain.cgraph.export_dendrogram(
			path, dpi, page_size, stroke_width = line_width
		)
		self.cmain.cview.set_recent_dir(path)
	
	
	def set_up_ExportCatalog(self, dialog):
		
		dialog.set_title("Export PDF Catalog")
		path = self.cmain.cview.get_recent_dir()
		if not path:
			path = self.cmain.cview.get_default_folder()
		filename = "%s.pdf" % sanitize_filename(
			self.cmain.cmodel.get_datasource_name() + "_catalog",
			"catalog",
		)
		frame = DialogExportDrawing(
			dialog,
			path = os.path.normpath(os.path.join(path, filename)),
			formats = "Adobe PDF (*.pdf)",
			multi = False,
			show_scale = True,
			show_dpi = False,
			show_page_size = True,
		)
		dialog.set_frame(frame)
	
	def process_ExportCatalog(self, dialog):
		
		path, scale, _, line_width, _, page_size = dialog.get_frame().get_data()
		if path is None:
			return
		
		if scale > 0:
			scale = 1 / scale
		else:
			scale = 1 / 3
		self.cmain.cgraph.export_catalog(
			path, scale, page_size, stroke_width = line_width
		)
		self.cmain.cview.set_recent_dir(path)
	
	
	def set_up_About(self, dialog):
		
		dialog.set_title("About CeraMatch")
		dialog.set_button_box(True, False)
		dialog.setModal(True)
		dialog.set_frame(DialogAbout())





from PySide2 import (QtWidgets, QtCore, QtGui)

class Button(QtWidgets.QPushButton):
	
	def __init__(self, caption, callback):
		
		QtWidgets.QPushButton.__init__(self, caption)
		
		self.clicked.connect(callback)



from PySide2 import (QtCore)

class CHistory(QtCore.QObject):
	
	def __init__(self, cmain):
		
		QtCore.QObject.__init__(self)
		
		self.cmain = cmain
		self._history = []
		self._index = 0
		self._paused = False
	
	def can_undo(self):
		
		return self._index > 0
	
	def can_redo(self):
		
		return self._index < len(self._history) - 1
	
	def clear(self):
		
		self._history.clear()
		self._index = 0
	
	def pause(self):
		
		self._paused = True
	
	def resume(self):
		
		self._paused = False
	
	def save(self):
		
		if self._paused:
			return
		
		data = self.cmain.cgraph.get_cluster_data()
		self._history = self._history[:self._index + 1]
		if self._history and (data == self._history[-1]):
			return
		
		self._history.append(data)
		self._index = len(self._history) - 1
		self.cmain.cactions.update()
	
	def undo(self):
		
		if self._index < 1:
			return
		self.pause()
		self._index -= 1
		self.cmain.cgraph.set_clusters(*self._history[self._index])
		self.resume()
		self.cmain.cactions.update()

	
	def redo(self):
		
		if self._index > len(self._history) - 1:
			return
		self.pause()
		self._index += 1
		self.cmain.cgraph.set_clusters(*self._history[self._index])
		self.resume()
		self.cmain.cactions.update()




from PySide2 import (QtWidgets, QtCore, QtGui)
from openpyxl import load_workbook
from pathlib import Path
import csv
import os

def get_columns_xlsx(path):
	
	wb = load_workbook(filename = path, read_only = True)
	columns = []
	for sheet in wb.sheetnames:
		ws = wb[sheet]
		for i, cell in enumerate(list(ws.iter_rows(max_row = 1))[0]):
			value = cell.value
			if value is not None:
				value = str(value).strip()
				if value:
					columns.append(value)
		break
	return columns

def get_columns_csv(path):
	
	columns = []
	with open(path, "r", newline = "") as f:
		reader = csv.reader(f, dialect = csv.excel, quoting=csv.QUOTE_ALL)
		for row in reader:
			for i, value in enumerate(row):
				value = value.strip()
				if value:
					columns.append(value)
			break
	return columns

class DialogImportClustering(QtWidgets.QFrame):
	
	def __init__(self, cview):
		
		QtWidgets.QFrame.__init__(self)
		
		self.cview = cview
		
		self.setMinimumWidth(512)
		self.setLayout(QtWidgets.QFormLayout())
		
		self.path_edit = QtWidgets.QLineEdit()
		self.path_edit.textChanged.connect(self.on_path_changed)
		path_button = QtWidgets.QPushButton("Browse")
		path_button.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
		path_button.clicked.connect(self.on_path_button)
		path_frame = QtWidgets.QWidget()
		path_frame.setLayout(QtWidgets.QHBoxLayout())
		path_frame.layout().setContentsMargins(0, 0, 0, 0)
		path_frame.layout().addWidget(self.path_edit)
		path_frame.layout().addWidget(path_button)
		
		self.sample_combo = QtWidgets.QComboBox()
		self.cluster_combo = QtWidgets.QComboBox()
		
		self.layout().addRow("Source File:", path_frame)
		self.layout().addRow("Sample Column:", self.sample_combo)
		self.layout().addRow("Cluster Column:", self.cluster_combo)
	
	@QtCore.Slot()
	def on_path_changed(self):
		
		path = self.path_edit.text()
		if not path:
			return
		if not os.path.isfile(path):
			return
		ext = os.path.splitext(path)[-1].lower()
		columns = []
		if ext == ".xlsx":
			columns = get_columns_xlsx(path)
		elif ext == ".csv":
			columns = get_columns_csv(path)
		if columns:
			self.sample_combo.clear()
			self.sample_combo.addItems(columns)
			self.cluster_combo.clear()
			self.cluster_combo.addItems(columns)
			if len(columns) > 1:
				self.cluster_combo.setCurrentIndex(1)
	
	@QtCore.Slot()
	def on_path_button(self):
		
		path, format = self.cview.get_load_path(
			"Select Source File",
			"Excel 2007+ Workbook (*.xlsx);;Comma-separated Values (*.csv)",
		)
		if path is None:
			return
		self.path_edit.setText(path)




from deposit_gui import DCActions

from deposit.utils.fnc_files import (as_url, sanitize_filename)

class CActions(DCActions):
	
	def __init__(self, cmain, cview) -> None:
		
		DCActions.__init__(self, cmain, cview)
		
		self.update()
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	
	
	# ---- Actions
	# ------------------------------------------------------------------------
	def set_up_tool_bar(self):
		
		return {
			"Data": [
				("Connect", "Connect"),
				("Save", "Save"),
				("Deposit", "Open Deposit"),
				None,
				("ImportClustering", "Import Clustering"),
				("ExportClustering", "Export Clustering"),
			],
			"Edit": [
				("Undo", "Undo"),
				("Redo", "Redo"),
				None,
				("SelectDescendants", "Select Descendants"),
			],
			"Export": [
				("ExportDendrogram", "Export Dendrogram as PDF"),
				("ExportCatalog", "Export Catalog as PDF"),
			],
		}
	
	def set_up_menu_bar(self):
		
		return {
			"Data": [
				("Connect", "Connect"),
				("Save", "Save"),
				("SaveAsFile", "Save As File"),
				("SaveAsPostgres", "Save As PostgreSQL"),
				("Deposit", "Open Deposit"),
				None,
				("AutoBackup", "Backup database after every save"),
				("ImportDescriptors", "Import Descriptors"),
				None,
				("ImportClustering", "Import Clustering"),
				("ExportClustering", "Export Clustering"),
			],
			"Edit": [
				("Undo", "Undo"),
				("Redo", "Redo"),
				None,
				("SelectDescendants", "Select Descendants"),
			],
			"Export": [
				("ExportDendrogram", "Export Dendrogram"),
				("ExportCatalog", "Export Catalog"),
			],
			"Help": [
				("About", "About"),
				("LogFile", "Log File"),
			],
		}
	
	
	# implement update_[name] and on_[name] for each action
	'''
	def update_[name](self):
		
		return dict(
			caption = "Caption",
			icon = "icon.svg",
			shortcut = "Ctrl+S",
			help = "Tool tip",
			combo: list,
			checkable = True,
			checked = True,
			enabled = True,
		)
	
	def on_[name](self, state):
		
		pass
	'''
	
	def update_Connect(self):
		
		return dict(
			caption = "Connect",
			icon = "connect.svg",
			help = "Connect to Database",
			checkable = False,
			enabled = True,
		)
	
	def on_Connect(self, state):
		
		self.cmain.cdialogs.open("Connect")
	
	
	def update_Save(self):
		
		return dict(
			caption = "Save",
			icon = "save.svg",
			help = "Save",
			checkable = False,
			enabled = not self.cmain.cmodel.is_saved(),
		)
	
	def on_Save(self, state):
		
		if self.cmain.cmodel.can_save():
			self.cmain.cmodel.save()
		else:
			self.on_SaveAsFile(True)
	
	def on_SaveAsFile(self, state):
		
		path, format = self.cmain.cview.get_save_path("Save Database As", "Pickle (*.pickle);;JSON (*.json)")
		if not path:
			return
		self.cmain.cview.set_recent_dir(path)
		self.cmain.cmodel.save(path = path)
		url = as_url(path)
		self.cmain.cdialogs.open("ConfirmLoad", url)
	
	
	def update_SaveAsPostgres(self):
		
		return dict(
			help = "Save As PostgreSQL Database",
			checkable = False,
			enabled = True,
		)
	
	def on_SaveAsPostgres(self, state):
		
		self.cmain.cdialogs.open("SaveAsPostgres")
	
	
	def update_Deposit(self):
		
		return dict(
			icon = "deposit.svg",
			help = "Open Deposit",
			checkable = False,
			enabled = True,
		)
	
	def on_Deposit(self, state):
		
		self.cmain.open_deposit()
	
	
	def update_ImportClustering(self):
		
		return dict(
			icon = "import_xls.svg",
			help = "Import Clustering",
			checkable = False,
			enabled = True,
		)
	
	def on_ImportClustering(self, state):
		
		self.cmain.cdialogs.open("ImportClustering")
	
	
	def update_ExportClustering(self):
		
		return dict(
			icon = "export_xls.svg",
			help = "Export Clustering",
			checkable = False,
			enabled = self.cmain.cgraph.has_clusters(),
		)
	
	def on_ExportClustering(self, state):
		
		filename = sanitize_filename(
			self.cmain.cmodel.get_datasource_name() + "_clusters.xlsx"
		)
		path, format = self.cmain.cview.get_save_path(
			"Export Clustering As",
			"Excel 2007+ Workbook (*.xlsx);;Comma-separated Values (*.csv)",
			filename = filename
		)
		if path is None:
			return
		self.cmain.cgraph.export_clusters(path, format)
	
	
	def update_Undo(self):
		
		return dict(
			caption = "Undo",
			icon = "undo.svg",
			help = "Undo last action (Ctrl+Z)",
			shortcut = "Ctrl+Z",
			checkable = False,
			enabled = self.cmain.history.can_undo(),
		)
	
	def on_Undo(self, state):
		
		self.cmain.history.undo()
	
	
	def update_Redo(self):
		
		return dict(
			caption = "Redo",
			icon = "redo.svg",
			help = "Redo last undone action (Ctrl+Y)",
			shortcut = "Ctrl+Y",
			checkable = False,
			enabled = self.cmain.history.can_redo(),
		)
	
	def on_Redo(self, state):
		
		self.cmain.history.redo()
	
	
	def update_SelectDescendants(self):
		
		return dict(
			caption = "Select Descendants (Ctrl+D)",
			icon = "select_descendants.svg",
			help = "Select Descendants of Node (Ctrl+D)",
			shortcut = "Ctrl+D",
			checkable = False,
			enabled = len(self.cmain.cgraph.get_selected()) > 0,
		)
	
	def on_SelectDescendants(self, state):
		
		self.cmain.cgraph.select_descendants()
	
	
	def update_ExportDendrogram(self):
		
		return dict(
			icon = "export_dendrogram.svg",
			help = "Export Dendrogram as PDF",
			checkable = False,
			enabled = self.cmain.cgraph.has_drawings(),
		)
	
	def on_ExportDendrogram(self, state):
		
		self.cmain.cdialogs.open("ExportDendrogram")
	
	
	def update_ExportCatalog(self):
		
		return dict(
			icon = "export_catalog.svg",
			help = "Export Clustered Drawings as PDF Catalog",
			checkable = False,
			enabled = self.cmain.cgraph.has_clusters(),
		)
	
	def on_ExportCatalog(self, state):
		
		self.cmain.cdialogs.open("ExportCatalog")
	
	
	def update_AutoBackup(self):
		
		return dict(
			help = "Backup database after every save",
			checkable = True,
			checked = self.cmain.cmodel.has_auto_backup(),
			enabled = True,
		)
	
	def on_AutoBackup(self, state):
		
		self.cmain.cmodel.set_auto_backup(state)
	
	
	def update_ImportDescriptors(self):
		
		return dict(
			help = "Import Descriptors from LAP application",
			checkable = False,
			enabled = True,
		)
	
	def on_ImportDescriptors(self, state):
		
		path, format = self.cmain.cview.get_load_path(
			"Import Descriptors",
			"Text file (*.txt)",
		)
		if not path:
			return
		self.cmain.import_descriptors(path)
	
	
	def on_About(self, state):
		
		self.cmain.cdialogs.open("About")
	
	def on_LogFile(self, state):
		
		self.cmain.open_in_external(self.cmain.cview.get_logging_path())




import multiprocessing as mp
import time

def _worker(worker_fnc, params_mp, collect_mp, max_queue_size, args):
	
	while True:
		try:
			params = params_mp.get()
		except:
			time.sleep(0.01)
			continue
		
		while collect_mp.qsize() > max_queue_size:
			time.sleep(0.01)
			pass
		collect_mp.put(worker_fnc(params, *args))

def process_mp(worker_fnc, params_list, worker_args = [], collect_fnc = None, collect_args = [], progress_fnc = None, progress_args = [], max_cpus = -1, max_queue_size = 100):
	
	def call_progress(progress_fnc, done, todo, progress_args):
		
		if progress_fnc == True:
			print("\r%d/%d             " % (done, todo), end = "")
		elif callable(progress_fnc):
			return progress_fnc(done, todo, *progress_args)
		return True
	
	params_mp = mp.Queue()
	for params in params_list:
		params_mp.put(params)
	todo = len(params_list)
	done = 0
	collect_mp = mp.Queue(todo)
	if max_cpus > 0:
		n_cpus = min(max_cpus, mp.cpu_count() - 1, todo)
	else:
		n_cpus = min(mp.cpu_count() - 1, todo)
	call_progress(progress_fnc, done, todo, progress_args)
	procs = []
	while done < todo:
		if len(procs) < n_cpus:
			procs.append(mp.Process(target = _worker, args = (worker_fnc, params_mp, collect_mp, max_queue_size, worker_args)))
			procs[-1].start()
		if not collect_mp.empty():
			data = collect_mp.get()
			done += 1
			if not call_progress(progress_fnc, done, todo, progress_args):
				break
			if collect_fnc is not None:
				collect_fnc(data, *collect_args)
		else:
			time.sleep(0.5)
	for proc in procs:
		proc.terminate()
		proc = None



from ceramatch.controller.controller import Controller
from deposit.utils.fnc_files import (clear_temp_dir)

from PySide2 import (QtWidgets)
import sys

class CMMain(object):
	
	def __init__(self):
		
		app = QtWidgets.QApplication(sys.argv)
#		app.setStyle("Fusion")
		
		self.controller = Controller()
		
		app.exec_()
	
	def __del__(self):
		
		clear_temp_dir(appdir = "ceramatch")



from ceramatch.view.vcontrols_groups.descriptors import Descriptors
from ceramatch.view.vcontrols_groups.distances import Distances
from ceramatch.view.vcontrols_groups.clusters import Clusters
from ceramatch.view.vcontrols_groups.attributes import Attributes
from ceramatch.view.vertical_scroll_area import VerticalScrollArea

from deposit.utils.fnc_files import (as_url, url_to_path)

from PySide2 import (QtWidgets, QtCore, QtGui)

class VControls(QtWidgets.QFrame):
	
	signal_folder_link_clicked = QtCore.Signal(str)		# path
	
	def __init__(self):
		
		QtWidgets.QFrame.__init__(self)
		
		self.descriptors = Descriptors()
		self.distances = Distances()
		self.clusters = Clusters()
		self.attributes = Attributes()
		
		self.setLayout(QtWidgets.QVBoxLayout())
		self.layout().setContentsMargins(10, 10, 0, 10)
		self.setSizePolicy(
			QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum
		)
		self.setMaximumWidth(400)
		
		self._dbname_label = QtWidgets.QLabel("Database: ")
		self._folder_label = QtWidgets.QLabel("Local Folder: ")
		self._folder_label.setWordWrap(True)
		self._folder_label.linkActivated.connect(self.on_folder_link)
		
		frame_db = QtWidgets.QFrame()
		frame_db.setLayout(QtWidgets.QVBoxLayout())
		frame_db.layout().setContentsMargins(5, 5, 5, 5)
		frame_db.setStyleSheet(
			".QFrame {border-style: solid; border-width: 1px; border-color: grey;}"
		)
		frame_db.layout().addWidget(self._dbname_label)
		frame_db.layout().addWidget(self._folder_label)
		
		frame_groups = QtWidgets.QFrame()
		frame_groups.setLayout(QtWidgets.QVBoxLayout())
		frame_groups.layout().setContentsMargins(0, 0, 0, 0)
		frame_groups.layout().addWidget(self.descriptors)
		frame_groups.layout().addWidget(self.distances)
		frame_groups.layout().addWidget(self.clusters)
		frame_groups.layout().addWidget(self.attributes)
		
		scroll_area = VerticalScrollArea(frame_groups)
		
		self.layout().addWidget(frame_db)
		self.layout().addWidget(scroll_area)
#		self.layout().addStretch()
	
	@QtCore.Slot(str)
	def on_folder_link(self, url):
		
		self.signal_folder_link_clicked.emit(url_to_path(url))
	
	def set_db_name(self, name):
		
		self._dbname_label.setText("Database: <b>%s</b>" % (name))
	
	def set_folder(self, path, url = None, max_path_length = 10):
		
		if path:
			if url is None:
				url = as_url(path)
			if len(path) > max_path_length:
				path = "\\ ".join(path.split("\\"))
			self._folder_label.setText("Local Folder: <a href=\"%s\">%s</a>" % (url, path))
		else:
			self._folder_label.setText("Local Folder: <b>not set</b>")




from ceramatch.view.vcontrols_groups.controls.button import (Button)

from PySide2 import (QtWidgets, QtCore, QtGui)

class Distances(QtWidgets.QGroupBox):
	
	signal_calculate = QtCore.Signal()
	signal_delete = QtCore.Signal()
	
	def __init__(self):
		
		QtWidgets.QGroupBox.__init__(self, "Distance")
		
		self.setStyleSheet("QGroupBox {font-weight: bold;}")
		self.setLayout(QtWidgets.QVBoxLayout())
		
		self.calculate_button = Button("Calculate Distances", self.on_calculate)
		self.delete_button = Button("Delete Distances", self.on_delete)
		
		self.layout().addWidget(self.calculate_button)
		self.layout().addWidget(self.delete_button)
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot()
	def on_calculate(self):
		
		self.signal_calculate.emit()
	
	@QtCore.Slot()
	def on_delete(self):
		
		self.signal_delete.emit()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def set_calculate_distances_enabled(self, state):
		
		self.calculate_button.setEnabled(state)
	
	def set_delete_distances_enabled(self, state):
		
		self.delete_button.setEnabled(state)




from deposit_gui import (DView, DNotification)

from ceramatch import __version__, __title__, res

from PySide2 import (QtWidgets, QtCore, QtGui)
import os

class View(DView):
	
	APP_NAME = __title__
	VERSION = __version__
	
	def __init__(self, vcontrols, vgraph) -> None:
		
		DView.__init__(self)
		
		self._close_callback = None
		
		self.set_res_folder(os.path.dirname(res.__file__))
		
		central_widget = QtWidgets.QWidget(self)
		central_widget.setLayout(QtWidgets.QVBoxLayout())
		central_widget.layout().setContentsMargins(0, 0, 0, 0)
		central_widget.layout().setSpacing(0)
		self.setCentralWidget(central_widget)
		
		self._tool_window = QtWidgets.QMainWindow()
		self._tool_window.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
		central_widget.layout().addWidget(self._tool_window)
		
		central_frame = QtWidgets.QFrame()
		central_frame.setLayout(QtWidgets.QHBoxLayout())
		central_frame.layout().setContentsMargins(0, 0, 0, 0)
		central_widget.layout().addWidget(central_frame)
		
		self.dummy_view = QtWidgets.QGraphicsView()
		self.dummy_view.hide()
		
		central_frame.layout().addWidget(vcontrols)
		central_frame.layout().addWidget(vgraph)
		central_frame.layout().addWidget(self.dummy_view)
		
		self._notification = DNotification(vgraph)
		
		self.setWindowIcon(self.get_icon("cm_icon.svg"))
	
	def show_notification(self, text, delay = None):
		
		self._notification.show(text, delay)
	
	def hide_notification(self):
		
		self._notification.hide()
	
	# events
	
	def exception_event(self, typ, value, tb):
		
		error_title = "%s: %s" % (str(typ), str(value)[:512])
		self.show_notification('''
Application Error: %s
(see Help -> Log File for details)
		''' % (error_title),
			delay = 7000,
		)
		text = "Exception: %s\nTraceback: %s" % (
			error_title, 
			"".join(traceback.format_tb(tb)),
		)
		self.logging.append(text)
		print(text)
	
	# overriden QMainWindow methods
	
	def addToolBar(self, title):
		
		return self._tool_window.addToolBar(title)
	
	def addToolBarBreak(self):
		
		self._tool_window.addToolBarBreak()
	
	def closeEvent(self, event):
		
		if self._close_callback is not None:
			if not self._close_callback():
				event.ignore()
				return
		DView.closeEvent(self, event)



from deposit_gui import AbstractSubcontroller
from deposit_gui.dgui.dgraph_view import DGraphView

from ceramatch.controller.cgraph_nodes.node import Node
from ceramatch.controller.cgraph_nodes.sample_node import SampleNode
from ceramatch.controller.cgraph_nodes.cluster_node import ClusterNode
from ceramatch.controller.cgraph_nodes.tool_tip import ToolTip
from ceramatch.utils.fnc_external import (
	save_xlsx, save_csv,
	import_clusters_xlsx, import_clusters_csv,
)
from ceramatch.utils.fnc_matching import (update_clusters)
from ceramatch.utils.fnc_drawing import (save_catalog)

from deposit.utils.fnc_files import (as_url)

from PySide2 import (QtCore, QtGui, QtWidgets)
import networkx as nx
import numpy as np
import copy
import os

class CGraph(AbstractSubcontroller):
	
	SCALE_DRAWINGS = 0.5
	SCALE_TOOLTIP = 1
	SCALE_CUTOFF = 0.2
	LINE_WIDTH = 1
	
	def __init__(self, cmain):
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._view = DGraphView()
		self._tool_tip = ToolTip()
		self._descendants = None
		
		self._view.signal_node_activated.connect(self.on_activated)
		self._view.signal_selected.connect(self.on_selected)
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot(object)
	def on_activated(self, node_id):
		
		node = self._view.get_node(node_id)
		if isinstance(node, SampleNode):
			self.cmain.open_deposit("SELECT [%(cls)s].* WHERE [%(cls)s] == %(obj_id)d" % dict(
				cls = self.cmain.cmodel.get_primary_class(),
				obj_id = node_id,
			))
		
		elif isinstance(node, ClusterNode):
			self.cmain.open_deposit("SELECT [%(cls)s].* WHERE [%(cls)s] == %(obj_id)d" % dict(
				cls = self.cmain.cmodel.get_cm_classes()["cluster"],
				obj_id = node_id,
			))
		
		else:
			self.select_descendants()
	
	@QtCore.Slot()
	def on_selected(self):
		
		self.cmain.ccontrols.update()
		
		self.cmain.ccontrols.set_attribute_data({})
		
		node_ids, _ = self._view.get_selected()
		if len(node_ids) == 1:
			node = self._view.get_node(node_ids[0])
			if isinstance(node, SampleNode):
				self.cmain.ccontrols.set_attribute_data(node.get_drawing_data())
		
		self.cmain.cactions.update()
	
	def on_hover(self, node, state):
		
		tool_tip = node.get_tool_tip()
		if tool_tip is None:
			self._tool_tip.hide()
		else:
			self._tool_tip.show(tool_tip, node)
	
	def on_moved(self, node):
		
		pos = node.pos()
		self.cmain.cmodel.save_position(node.node_id, pos.x(), pos.y())
		self.cmain.history.save()
	
	def on_mouse_released(self, node):
		
		pass
	
	def on_drop(self, src_node, tgt_node):
		
		if isinstance(src_node, SampleNode) and isinstance(tgt_node, ClusterNode):
			node_ids, _ = self._view.get_selected()
			src_nodes = []
			for node_id in node_ids:
				node = self._view.get_node(node_id)
				if not isinstance(node, SampleNode):
					continue
				src_nodes.append(node)
			if not src_nodes:
				return
			
			self.add_to_cluster(src_nodes, tgt_node)
			self.cmain.history.save()
		
		if isinstance(src_node, ClusterNode):
			if isinstance(tgt_node, ClusterNode):
				self.merge_clusters(src_node, tgt_node)
				self.cmain.history.save()
			elif tgt_node.__class__ == Node:
				self.reparent_cluster(src_node, tgt_node)
				self.cmain.history.save()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def get_selected_samples(self):
		
		node_ids, _ = self._view.get_selected()
		selected = []
		for node_id in node_ids:
			node = self._view.get_node(node_id)
			if not isinstance(node, SampleNode):
				continue
			selected.append(node_id)
		
		return selected
	
	def get_selected_clusters(self):
		
		node_ids, _ = self._view.get_selected()
		selected = []
		for node_id in node_ids:
			node = self._view.get_node(node_id)
			if not isinstance(node, ClusterNode):
				continue
			selected.append(node_id)
		
		return selected
	
	def get_selected(self):
		
		node_ids, _ = self._view.get_selected()
		
		return node_ids
	
	def get_scale_factor(self):
		
		return self._view.get_scale_factor()
	
	def clear(self):
		
		self._descendants = None
		self._view.clear()
	
	def populate(self, drawing_data, cluster_data, node_data, edges):
		# drawing_data = {obj_id: {key: value, key: [{key: value, ...}, ...], ...}, ...}
		# cluster_data = {obj_id: {"name": value, "position": (x, y), "children": [obj_id, ...], ...}, ...}
		# node_data = {obj_id: {"name": value, "position": (x, y), ...}, ...}
		# edges = [(source_id, target_id), ...]
		
		nodes = []  # [SampleNode, ClusterNode, ...]
		positions = {}  # {node_id: (x, y), ...}
		move_nodes = []  # [node_id, ...]; if set, only calculate positions for those nodes
		
		self._descendants = None
		
		# prune clusters & nodes
		collect = {}
		for obj_id in cluster_data:
			if cluster_data[obj_id]["children"]:
				collect[obj_id] = cluster_data[obj_id]
		cluster_data = collect
		G = nx.DiGraph()
		for source_id, target_id in edges:
			G.add_edge(source_id, target_id)
		collect = {}
		cluster_ids = set(cluster_data.keys())
		for obj_id in node_data:
			if cluster_ids.intersection(nx.descendants(G, obj_id)):
				collect[obj_id] = node_data[obj_id]
		node_data = collect
		
		self.cmain.cview.progress.show("Rendering Drawings")
		self.cmain.cview.progress.update_state(value = 0, maximum = len(drawing_data))
		cnt = 1
		for obj_id in drawing_data:
			self.cmain.cview.progress.update_state(value = cnt)
			if self.cmain.cview.progress.cancel_pressed():
				return
			cnt += 1
			nodes.append(SampleNode(self,
				node_id = obj_id,
				label = drawing_data[obj_id][self.cmain.cmodel.NAME_ID],
				drawing_data = drawing_data[obj_id],
			))
			pos = drawing_data[obj_id].get("position")
			if pos is not None:
				positions[obj_id] = pos
		
		for obj_id in cluster_data:
			nodes.append(ClusterNode(self,
				node_id = obj_id,
				label = cluster_data[obj_id]["name"],
				children = cluster_data[obj_id]["children"],
			))
			positions[obj_id] = cluster_data[obj_id]["position"]
		
		for obj_id in node_data:
			nodes.append(Node(self,
				node_id = obj_id,
				label = node_data[obj_id]["name"],
			))
			positions[obj_id] = node_data[obj_id]["position"]
		
		self.cmain.cview.set_dummy_graph(True)
		
		self._view.populate(
			nodes, edges, positions,
			x_stretch = 1,
			y_stretch = 2,
			edge_arrow_size = 0,
			move_nodes = move_nodes,
			progress = self.cmain.cview.progress,
		)
		
		for node in self._view.get_nodes():
			node._moved = False
		
		self.cmain.cview.set_dummy_graph(False)
	
	def set_clusters(self, objects, clusters, node_idxs, edges, labels, positions = {}):
		# objects = [obj_id, ...]
		# clusters = {node_idx: [object_idx, ...], ...}; object_idx = index in objects
		# node_idxs = [node_idx, ...]
		# edges = [(source_idx, target_idx), ...]
		# labels = {node_idx: label, ...}
		# positions = {node_idx: (x, y), ...}
		
		self.cmain.cmodel._model.blockSignals(True)
		
		added_ids = set()
		changed_ids = set()
		node_id_lookup = {}
		self._descendants = None
		
		nodes = []
		for node in self._view.get_nodes():
			if not isinstance(node, SampleNode):
				continue
			if node.node_id not in objects:
				continue
			nodes.append(node.copy())
			node_id_lookup[objects.index(node.node_id)] = node.node_id
		
		for node_idx in list(clusters.keys()):
			children = [objects[idx] for idx in clusters[node_idx] if idx in node_id_lookup]
			if not children:
				del clusters[node_idx]
				continue
			node_id = self.cmain.cmodel.add_cluster(labels[node_idx], children)
			added_ids.add(node_id)
			node_id_lookup[node_idx] = node_id
			nodes.append(ClusterNode(self,
				node_id = node_id,
				label = labels[node_idx],
				children = children,
			))
		
		G = nx.DiGraph()
		for source_idx, target_idx in edges:
			G.add_edge(source_idx, target_idx)
		cluster_idxs = set(clusters.keys())
		
		for node_idx in node_idxs:
			if node_idx in node_id_lookup:
				continue
			if node_idx < len(objects):
				continue
			if not cluster_idxs.intersection(nx.descendants(G, node_idx)):
				continue
			node_id = self.cmain.cmodel.add_node(labels[node_idx])
			added_ids.add(node_id)
			node_id_lookup[node_idx] = node_id
			nodes.append(Node(self,
				node_id = node_id,
				label = labels[node_idx],
			))
		
		edges_ = []
		for source_idx, target_idx in edges:
			if (source_idx not in node_id_lookup) or (target_idx not in node_id_lookup):
				continue
			source_id = node_id_lookup[source_idx]
			target_id = node_id_lookup[target_idx]
			edges_.append([source_id, target_id])
			self.cmain.cmodel.add_link(source_id, target_id)
			changed_ids.add(source_id)
			changed_ids.add(target_id)
		
		self.cmain.cview.set_dummy_graph(True)
		
		self._view.populate(
			nodes = nodes, 
			edges = edges_, 
			positions = dict([(node_id_lookup[node_idx], positions[node_idx]) \
				for node_idx in positions]),
			x_stretch = 1,
			y_stretch = 1 if positions else 2,
			edge_arrow_size = 0,
			progress = self.cmain.cview.progress,
		)
		
		self.cmain.cview.set_dummy_graph(False)
		
		for node in self._view.get_nodes():
			pos = node.pos()
			self.cmain.cmodel.save_position(node.node_id, pos.x(), pos.y())
			node._moved = False
		
		self.cmain.cmodel._model.blockSignals(False)
		self.cmain.cmodel._model.on_added(added_ids)
		self.cmain.cmodel._model.on_changed(changed_ids)
	
	def get_cluster_data(self):
		# returns objects, clusters, node_idxs, edges, labels, positions
		#	objects = [obj_id, ...]
		#	clusters = {node_idx: [object_idx, ...], ...}; object_idx = index in objects
		#	node_idxs = [node_idx, ...]
		#	edges = [(source_idx, target_idx), ...]
		#	labels = {node_idx: label, ...}
		#	positions = {node_idx: (x, y), ...}
		
		objects = [obj.id for obj in self.cmain.cmodel.get_drawing_objects()]
		nodes = objects.copy()
		for node in self._view.get_nodes():
			if node.node_id not in nodes:
				nodes.append(node.node_id)
		node_idxs = list(range(len(nodes)))
		clusters = self.get_clusters()
		edges = []
		for edge in self._view.get_edges():
			edges.append((
				nodes.index(edge.source().node_id), 
				nodes.index(edge.target().node_id),
			))
		labels = {}
		positions = {}
		for node in self._view.get_nodes():
			node_idx = nodes.index(node.node_id)
			labels[node_idx] = node.label
			pos = node.pos()
			positions[node_idx] = (pos.x(), pos.y())
		clusters = dict([(
			nodes.index(node_id),
			[nodes.index(sample_id) for sample_id in clusters[node_id]],
		) for node_id in clusters])
		
		return objects, clusters, node_idxs, edges, labels, positions
	
	def get_descendants(self, node_id):
		
		if self._descendants is None:
			G = nx.DiGraph()
			for edge in self._view.get_edges():
				G.add_edge(edge.source().node_id, edge.target().node_id)
			self._descendants = {}
			for node_id_ in G.nodes():
				self._descendants[node_id_] = set(nx.descendants(G, node_id_))
		
		return self._descendants[node_id]
	
	def select_descendants(self):
		
		node_ids = set()
		for node_id in self.get_selected():
			node_ids.update(self.get_descendants(node_id))
		cnt = 1
		self._view.blockSignals(True)
		for node_id in node_ids:
			cnt += 1
			self._view.select_node(node_id)
		self._view.blockSignals(False)
	
	def del_node(self, node_id):
		
		self.cmain.cmodel.del_object(node_id)
		self._view.del_node(node_id)
	
	def update_sample_data(self, node_id, data):
		# data = {name: value, ...}
		
		node = self._view.get_node(node_id)
		if not isinstance(node, SampleNode):
			return
		drawing_data = node.get_drawing_data()
		drawing_data.update(data)
		node.set_drawing_data(data)
	
	def add_to_cluster(self, src_nodes, tgt_node):
		
		self._view.del_edges([node.node_id for node in src_nodes])
		
		for src_node in src_nodes:
			self._view.add_edge(
				tgt_node.node_id, src_node.node_id,
				color = QtCore.Qt.black,
				arrow_size = 0,
			)
			self.cmain.cmodel.set_cluster(src_node.node_id, tgt_node.node_id)
		
		self.prune_tree()
	
	def merge_clusters(self, src_node, tgt_node):
		
		src_node = self._view.get_node(src_node.node_id)
		tgt_node = self._view.get_node(tgt_node.node_id)
		src_nodes = src_node.get_children()
		self.del_node(src_node.node_id)
		self.add_to_cluster(src_nodes, tgt_node)
	
	def reparent_cluster(self, src_node, tgt_node):
		
		self.cmain.cmodel._model.blockSignals(True)
		changed_ids = set()
		
		del_links = set()
		for parent in src_node.get_parents():
			del_links.add((parent.node_id, src_node.node_id))
		for src_id, tgt_id in del_links:
			self._view.del_edge(src_id, tgt_id)
			self.cmain.cmodel.del_link(src_id, tgt_id)
			changed_ids.add(src_id)
			changed_ids.add(tgt_id)
		self._view.add_edge(
			tgt_node.node_id,
			src_node.node_id,
			color = QtCore.Qt.black,
			arrow_size = 0,
		)
		self.cmain.cmodel.add_link(src_node.node_id, tgt_node.node_id)
		changed_ids.add(src_node.node_id)
		changed_ids.add(tgt_node.node_id)
		
		self.cmain.cmodel._model.blockSignals(False)
		self.cmain.cmodel._model.on_changed(changed_ids)
		
		self.prune_tree()
	
	def add_manual_cluster(self):
		
		sample_ids = self.get_selected_samples()
		if not sample_ids:
			return
		
		self.cmain.cmodel._model.blockSignals(True)
		changed_ids = set(sample_ids)
		
		x = 0
		y = np.inf
		parent_ids = set()
		for node_id in sample_ids:
			node = self._view.get_node(node_id)
			pos = node.pos()
			x += pos.x()
			y = min(y, pos.y())
			to_del = []
			for node_ in node.get_parents():
				for parent in node_.get_parents():
					parent_ids.add(parent.node_id)
				self._view.del_edge(node_.node_id, node_id)
		parent_id = None
		if parent_ids:
			parent_id = min(parent_ids)
		x /= len(sample_ids)
		y -= 10
		label = "Manual_%d" % (node_id)
		cluster_id = self.cmain.cmodel.add_cluster(label, sample_ids)
		cluster = ClusterNode(self,
			node_id = cluster_id,
			label = label,
			children = sample_ids,
		)
		self._view.add_node(cluster, (x, y))
		for sample_id in sample_ids:
			self._view.add_edge(cluster_id, sample_id, color = QtCore.Qt.black, arrow_size = 0)
		if parent_id is not None:
			self._view.add_edge(parent_id, cluster_id, color = QtCore.Qt.black, arrow_size = 0)
			self.cmain.cmodel.add_link(parent_id, cluster_id)
		
		self.cmain.cmodel._model.blockSignals(False)
		self.cmain.cmodel._model.on_changed(changed_ids)
		self.cmain.cmodel._model.on_added([cluster_id])
		
		self.prune_tree()
	
	def rename_cluster(self):
		
		obj_id = self.get_selected_clusters()
		if len(obj_id) != 1:
			return
		obj_id = obj_id[0]
		node = self._view.get_node(obj_id)
		label = self.cmain.cview.show_input_dialog("Rename Cluster", "Label:", node.label)
		node.set_label(label)
		self.cmain.cmodel.set_cluster_label(obj_id, label)
	
	def prune_tree(self):
		
		self.cmain.cmodel._model.blockSignals(True)
		deleted_ids = set()
		
		to_del = []
		for node in self._view.get_nodes():
			if not (isinstance(node, ClusterNode) or node.__class__ == Node):
				continue
			if not node.has_child():
				to_del.append(node)
		for node in to_del:
			deleted_ids.add(node.node_id)
			self.del_node(node.node_id)
		
		self.cmain.cmodel._model.blockSignals(False)
		self.cmain.cmodel._model.on_deleted(deleted_ids)
	
	def export_clusters(self, path, format):
		
		header = ["Sample ID", "Cluster"]
		rows = []
		for node in self._view.get_nodes():
			if not isinstance(node, ClusterNode):
				continue
			cluster = node.label
			for sample in node.get_children():
				rows.append([sample.label, cluster])
		
		if format == "Excel 2007+ Workbook (*.xlsx)":
			save_xlsx(header, rows, path)
		
		elif format == "Comma-separated Values (*.csv)":
			save_csv(header, rows, path)
		
		self.cmain.cview.show_notification(
			'''
Exported to: <a href="%s">%s</a>
			''' % (as_url(path), path),
			delay = 7000,
		)
	
	def import_clusters(self, path, sample_column, cluster_column):
		
		format = os.path.splitext(path)[-1].strip(".").lower()
		clusters = {}
		if format == "xlsx":
			clusters = import_clusters_xlsx(path, sample_column, cluster_column)
		elif format == "csv":
			clusters = import_clusters_csv(path, sample_column, cluster_column)
		if not clusters:
			return
		
		sample_id_lookup = {}
		for node in self._view.get_nodes():
			if not isinstance(node, SampleNode):
				continue
			sample_id_lookup[node.label] = node.node_id
		collect = {}
		labels = {}
		idx = 0
		for cluster_label in clusters:
			samples = []
			for label in clusters[cluster_label]:
				if label not in sample_id_lookup:
					continue
				samples.append(sample_id_lookup[label])
			if samples:
				collect[idx] = samples.copy()
				labels[idx] = cluster_label
				idx += 1
		clusters = collect
		if not clusters:
			return		
		
		self.cmain.cview.progress.show("Clustering")
		objects, clusters, nodes, edges, labels = update_clusters(
			self.cmain.cmodel._distance,
			clusters = clusters,
			labels = labels,
			progress = self.cmain.cview.progress,
		)
		self.cmain.cview.progress.stop()
		
		self.cmain.cmodel.delete_clusters()
		self.set_clusters(objects, clusters, nodes, edges, labels)
		self.cmain.history.save()
		self.cmain.ccontrols.update()
	
	def export_dendrogram(self, path, dpi, page_size, stroke_width):
		
		scale = self._view.get_scale_factor()
		self._view.scale_view(1/scale)
		self._view.save_pdf(path, dpi, page_size, stroke_width)
		self._view.scale_view(scale)
		self.cmain.cview.show_notification(
			'''
Dendrogram exported to: <a href="%s">%s</a>
			''' % (as_url(path), path),
			delay = 7000,
		)
	
	def export_catalog(self, path, scale, page_size, stroke_width):
		
		self.cmain.cview.progress.show("Saving Catalog")
		
		data = {}
		clusters = {}
		labels = {}
		for node in self._view.get_nodes():
			labels[node.node_id] = node.label
			if isinstance(node, SampleNode):
				data[node.node_id] = node.get_drawing_data()
			elif isinstance(node, ClusterNode):
				clusters[node.node_id] = [
					child.node_id for child in node.get_children()
				]
		# data = {sample_id: {key: value, key: [{key: value, ...}, ...], ...}, ...}
		# clusters = {obj_id: [sample_id, ...], ...}
		# labels = {obj_id: label, ...}
		
		save_catalog(
			path, data, clusters, labels, scale, page_size, stroke_width, 
			progress = self.cmain.cview.progress,
		)
		self.cmain.cview.progress.stop()
		
		self.cmain.cview.show_notification(
			'''
Catalog exported to: <a href="%s">%s</a>
			''' % (as_url(path), path),
			delay = 7000,
		)
	
	def has_drawings(self):
		
		return self._view.has_nodes()
	
	def has_clusters(self):
		
		for node in self._view.get_nodes():
			if not isinstance(node, SampleNode):
				return True
		return False
	
	def get_n_clusters(self):
		
		n_clusters = 0
		for node in self._view.get_nodes():
			if isinstance(node, ClusterNode):
				n_clusters += 1
		
		return n_clusters
	
	def get_clusters(self):
		# returns clusters = {node_idx: [obj_id, ...], ...}
		
		clusters = {}
		for node in self._view.get_nodes():
			if not isinstance(node, ClusterNode):
				continue
			clusters[node.node_id] = [
				child.node_id for child in node.get_children()
			]
		return clusters
	
	def get_cluster_labels(self):
		
		labels = {}
		for node in self._view.get_nodes():
			if not isinstance(node, ClusterNode):
				continue
			labels[node.node_id] = node.label
		return labels
	
	def get_profile_data(self):
		# returns {obj_id: (coords, radius), ...}
		
		data = {}
		for node in self._view.get_nodes():
			if not isinstance(node, SampleNode):
				continue
			coords, radius = node.get_profile_data()
			if radius is None:
				continue
			data[node.node_id] = (coords, radius)
		
		return data





from PySide2 import (QtWidgets, QtCore, QtGui)

class LineEdit(QtWidgets.QLineEdit):
	
	def __init__(self, default = ""):
		
		QtWidgets.QLineEdit.__init__(self, default)
		
		self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)




from ceramatch.view.vcontrols_groups.controls.button import (Button)
from ceramatch.view.vcontrols_groups.controls.combo import (Combo)
from ceramatch.view.vcontrols_groups.controls.line_edit import (LineEdit)

from PySide2 import (QtWidgets, QtCore, QtGui)

class Clusters(QtWidgets.QGroupBox):
	
	signal_cluster = QtCore.Signal()
	signal_update_tree = QtCore.Signal()
	signal_add_cluster = QtCore.Signal()
	signal_rename_cluster = QtCore.Signal()
	signal_delete = QtCore.Signal()
	signal_n_clusters_changed = QtCore.Signal()
	
	def __init__(self):
		
		QtWidgets.QGroupBox.__init__(self, "Clustering")
		
		self.setStyleSheet("QGroupBox {font-weight: bold;}")
		self.setLayout(QtWidgets.QVBoxLayout())
		
		self.cluster_button = Button("Auto Cluster", self.on_cluster)
		self.n_clusters_combo = Combo(self.on_n_clusters_changed)
		self.limit_edit = LineEdit("0.68")
		self.n_samples_label = QtWidgets.QLabel("")
		self.n_clusters_label = QtWidgets.QLabel("")
		self.update_tree_button = Button("Update Tree", self.on_update_tree)
		self.add_cluster_button = Button("Make Cluster", self.on_add_cluster)
		self.rename_cluster_button = Button("Rename Cluster", self.on_rename_cluster)
		self.delete_button = Button("Clear Clusters", self.on_delete)
		
		form_frame = QtWidgets.QFrame()
		form_frame.setLayout(QtWidgets.QFormLayout())
		form_frame.layout().setContentsMargins(0, 0, 0, 0)
		form_frame.layout().addRow(
			QtWidgets.QLabel("N Clusters:"), self.n_clusters_combo
		)
		form_frame.layout().addRow(
			QtWidgets.QLabel("Dist. Limit:"), self.limit_edit
		)
		form_frame.layout().addRow(
			QtWidgets.QLabel("Samples:"), self.n_samples_label
		)
		form_frame.layout().addRow(
			QtWidgets.QLabel("Clusters Found:"), self.n_clusters_label
		)
		
		self.layout().addWidget(self.cluster_button)
		self.layout().addWidget(form_frame)
		self.layout().addWidget(self.update_tree_button)
		self.layout().addWidget(self.add_cluster_button)
		self.layout().addWidget(self.rename_cluster_button)
		self.layout().addWidget(self.delete_button)
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot()
	def on_cluster(self):
		
		self.signal_cluster.emit()
	
	@QtCore.Slot()
	def on_update_tree(self):
		
		self.signal_update_tree.emit()
	
	@QtCore.Slot()
	def on_add_cluster(self):
		
		self.signal_add_cluster.emit()
	
	@QtCore.Slot()
	def on_rename_cluster(self):
		
		self.signal_rename_cluster.emit()
	
	@QtCore.Slot()
	def on_delete(self):
		
		self.signal_delete.emit()
	
	@QtCore.Slot()
	def on_n_clusters_changed(self):
		
		self.signal_n_clusters_changed.emit()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def get_limits(self):
		
		try:
			n_clusters = int(self.n_clusters_combo.get_value())
		except:
			n_clusters = None
		try:
			limit = float(self.limit_edit.text())
		except:
			limit = None
		return n_clusters, limit
	
	def set_clustering_enabled(self, state):
		
		self.cluster_button.setEnabled(state)
		self.n_clusters_combo.setEnabled(state)
		self.limit_edit.setEnabled(state)
	
	def set_limit_enabled(self, state):
		
		self.limit_edit.setEnabled(state)
	
	def set_update_tree_enabled(self, state):
		
		self.update_tree_button.setEnabled(state)
	
	def set_add_cluster_enabled(self, state):
		
		self.add_cluster_button.setEnabled(state)
	
	def set_rename_cluster_enabled(self, state):
		
		self.rename_cluster_button.setEnabled(state)
	
	def set_delete_enabled(self, state):
		
		self.delete_button.setEnabled(state)
	
	def set_n_clusters(self, values, n_clusters = None):
		
		self.n_clusters_combo.set_values(values, n_clusters)
	
	def set_n_samples(self, n_samples):
		
		if n_samples is None:
			n_samples = ""
		self.n_samples_label.setText(str(n_samples))
	
	def set_n_found(self, n_found):
		
		if n_found is None:
			n_found = ""
		self.n_clusters_label.setText(str(n_found))




from PySide2 import (QtWidgets, QtCore, QtGui)

class VerticalScrollArea(QtWidgets.QScrollArea):
	
	def __init__(self, contents):
		
		self.contents = contents
		
		QtWidgets.QScrollArea.__init__(self)
		
		self.setWidgetResizable(True)
		self.setFrameStyle(QtWidgets.QFrame.NoFrame)
		self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
		self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
		self.setWidget(self.contents)



from deposit.utils.fnc_serialize import (try_numeric)

from PySide2 import (QtWidgets, QtCore, QtGui)
from natsort import natsorted
import numbers
import os

class DialogExportDrawing(QtWidgets.QFrame):
	
	def __init__(
		self, dialog, path, formats, 
		multi = False, show_scale = True, show_dpi = True, show_page_size = False
	):
		QtWidgets.QFrame.__init__(self)
		
		self.formats = formats
		self.multi = multi
		self.show_dpi = show_dpi
		self.show_scale = show_scale
		self.show_page_size = show_page_size
		if os.path.isdir(path):
			self.folder = path
		else:
			self.folder = os.path.dirname(path)
		
		self.page_sizes = {}
		for name in natsorted(QtGui.QPageSize.PageSizeId.values):
			self.page_sizes[name] = QtGui.QPageSize.PageSizeId.__dict__[name]
		
		dialog.setModal(True)
		dialog.set_button_box(True, True)
		dialog.set_enabled(True)
		
		self.setMinimumWidth(256)
		self.setLayout(QtWidgets.QFormLayout())
		
		if self.show_scale:
			self.scale_edit = QtWidgets.QLineEdit()
			self.scale_edit.setPlaceholderText("Auto")
			self.scale_edit.setFixedWidth(40)
			self.scale_frame = QtWidgets.QWidget()
			self.scale_frame.setLayout(QtWidgets.QHBoxLayout())
			self.scale_frame.layout().setContentsMargins(0, 0, 0, 0)
			self.scale_frame.layout().addWidget(QtWidgets.QLabel("1:"))
			self.scale_frame.layout().addWidget(self.scale_edit)
			self.scale_frame.layout().addStretch()
		
		if self.show_dpi:
			self.dpi_edit = QtWidgets.QLineEdit("600")
			self.dpi_edit.setFixedWidth(40)
		
		if self.show_page_size:
			self.page_size_combo = QtWidgets.QComboBox()
			self.page_size_combo.setEditable(False)
			items = list(self.page_sizes.keys())
			self.page_size_combo.addItems(items)
			self.page_size_combo.setCurrentIndex(items.index("A4"))
		
		self.linewidth_edit = QtWidgets.QLineEdit("0.2")
		self.linewidth_edit.setFixedWidth(40)
		
		if self.multi and self.formats:
			self.format_combo = QtWidgets.QComboBox()
			self.format_combo.setEditable(False)
			self.format_combo.addItems(list(self.formats.keys()))
		
		self.dest_edit = QtWidgets.QLineEdit()
		self.dest_button = QtWidgets.QPushButton("Browse")
		self.dest_button.clicked.connect(self.on_dest_button)
		self.dest_frame = QtWidgets.QWidget()
		self.dest_frame.setLayout(QtWidgets.QHBoxLayout())
		self.dest_frame.layout().setContentsMargins(0, 0, 0, 0)
		self.dest_frame.layout().addWidget(self.dest_edit)
		self.dest_frame.layout().addWidget(self.dest_button)
		
		if self.show_scale:
			self.layout().addRow("Scale:", self.scale_frame)
		if self.show_dpi:
			self.layout().addRow("DPI:", self.dpi_edit)
		if self.show_page_size:
			self.layout().addRow("Page Size:", self.page_size_combo)
		self.layout().addRow("Line Width:", self.linewidth_edit)
		if self.multi and self.formats:
			self.layout().addRow("Format:", self.format_combo)
		self.layout().addRow("Destination:", self.dest_frame)
		
		if multi:
			self.dest_edit.setText(self.folder)
		else:
			self.dest_edit.setText(path)
	
	@QtCore.Slot()
	def on_dest_button(self):
		
		path = self.dest_edit.text()
		if self.multi:
			path = QtWidgets.QFileDialog.getExistingDirectory(None,
				caption = "Select Destination Folder", dir = self.folder
			)
		else:
			path, _ = QtWidgets.QFileDialog.getSaveFileName(None, 
				"Select Destination File", 
				path, 
				self.formats,
			)
		if path:
			self.dest_edit.setText(os.path.normpath(path))
	
	def get_data(self):
		
		path = self.dest_edit.text().strip()
		scale = 0
		if self.show_scale:
			scale = try_numeric(self.scale_edit.text().strip())
			if not scale:
				scale = 0
		dpi = 0
		if self.show_dpi:
			dpi = try_numeric(self.dpi_edit.text().strip())
		line_width = try_numeric(self.linewidth_edit.text().strip())
		page_size = QtGui.QPageSize.A4
		if self.show_page_size:
			page_size = self.page_sizes[self.page_size_combo.currentText()]
		format = None
		if self.multi and self.formats:
			format = self.formats[self.format_combo.currentText()]
		for value in [scale, dpi, line_width]:
			if not isinstance(value, numbers.Number):
				return None, None, None, None, None
		
		return path, scale, dpi, line_width, format, page_size




from deposit_gui import AbstractSubcontroller

from ceramatch.view.vcontrols import VControls

from ceramatch.utils.fnc_matching import (get_clusters, update_clusters)

from PySide2 import (QtCore)

class CControls(AbstractSubcontroller):
	
	def __init__(self, cmain) -> None:
		
		AbstractSubcontroller.__init__(self, cmain)
		
		self._view = VControls()
		
		self._view.signal_folder_link_clicked.connect(self.on_folder_link_clicked)
		
		self._view.descriptors.signal_load_drawings.connect(self.on_load_drawings)
		self._view.descriptors.signal_cluster_classes_changed.connect(
			self.on_cluster_classes_changed
		)
		
		self._view.distances.signal_calculate.connect(self.on_calculate_distances)
		self._view.distances.signal_delete.connect(self.on_delete_distances)
		
		self._view.clusters.signal_cluster.connect(self.on_cluster)
		self._view.clusters.signal_update_tree.connect(self.on_update_tree)
		self._view.clusters.signal_add_cluster.connect(self.on_add_cluster)
		self._view.clusters.signal_rename_cluster.connect(self.on_rename_cluster)
		self._view.clusters.signal_delete.connect(self.on_delete_clusters)
		self._view.clusters.signal_n_clusters_changed.connect(self.on_n_clusters_changed)
		self._view.attributes.signal_store_attributes.connect(self.on_store_attributes)
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	def on_loaded(self):
		
		descriptors = self.cmain.cmodel.get_descriptors()
		attributes = self.cmain.cmodel.get_attributes()
		cls = self.cmain.cmodel.get_primary_class()
		classes = self.cmain.cmodel.get_class_names(ordered = True)
		
		cm_classes = self.cmain.cmodel.get_cm_classes()
		
		descriptors_ = {}  # {label: chain, ...}
		name_lookup = dict([(name, label) for label, _, name in attributes])
		for name, chain in descriptors:
			descriptors_[name_lookup.get(name, name)] = chain
		
		self._view.descriptors.set_descriptors(descriptors_)
		self._view.descriptors.set_cluster_class(cm_classes["cluster"], classes)
		self._view.descriptors.set_node_class(cm_classes["node"], classes)
		self._view.descriptors.set_position_class(cm_classes["position"], classes)
		
		self.update()
	
	@QtCore.Slot(str)
	def on_folder_link_clicked(self, path):
		
		self.cmain.open_folder(path)
	
	@QtCore.Slot()
	def on_load_drawings(self):
		
		self.cmain.history.clear()
		self.cmain.cmodel.load_drawings()
		self.cmain.history.save()
		self.update()
	
	@QtCore.Slot()
	def on_cluster_classes_changed(self):
		
		self.cmain.cmodel.set_cm_classes(dict(
			cluster = self._view.descriptors.get_cluster_class(),
			node = self._view.descriptors.get_node_class(),
			position = self._view.descriptors.get_position_class(),
		))
	
	@QtCore.Slot()
	def on_calculate_distances(self):
		
		self.cmain.cdialogs.open("SelectDistances")
	
	@QtCore.Slot()
	def on_delete_distances(self):
		
		if not self.cmain.cview.show_question(
			"Delete Distances",
			"Are you sure to delete distances from database?\nWarning: This action cannot be undone!",
		):
			return
		self.cmain.cmodel.delete_distance()
		self.update()
	
	@QtCore.Slot()
	def on_cluster(self):
		
		n_clusters, limit = self._view.clusters.get_limits()
		self.cmain.cview.progress.show("Clustering")
		objects, clusters, nodes, edges, labels = get_clusters(
			self.cmain.cmodel._distance,
			max_clusters = n_clusters,
			limit = limit,
			progress = self.cmain.cview.progress,
		)
		self.cmain.cview.progress.stop()
		
		self.cmain.cmodel.delete_clusters()
		self.cmain.cgraph.set_clusters(objects, clusters, nodes, edges, labels)
		self.cmain.history.save()
		
		self.update()
	
	@QtCore.Slot()
	def on_update_tree(self):
		
		self.cmain.cview.progress.show("Clustering")
		objects, clusters, nodes, edges, labels = update_clusters(
			self.cmain.cmodel._distance,
			clusters = self.cmain.cgraph.get_clusters(),
			labels = self.cmain.cgraph.get_cluster_labels(),
			progress = self.cmain.cview.progress,
		)
		self.cmain.cview.progress.stop()
		
		self.cmain.cmodel.delete_clusters()
		self.cmain.cgraph.set_clusters(objects, clusters, nodes, edges, labels)
		self.cmain.history.save()
		
		self.update()
	
	@QtCore.Slot()
	def on_add_cluster(self):
		
		self.cmain.cgraph.add_manual_cluster()
		self.cmain.history.save()
	
	@QtCore.Slot()
	def on_rename_cluster(self):
		
		self.cmain.cgraph.rename_cluster()
		self.cmain.history.save()
	
	@QtCore.Slot()
	def on_delete_clusters(self):
		
		self.cmain.cmodel.delete_clusters()
		self.cmain.history.save()
		self.update()
	
	@QtCore.Slot()
	def on_n_clusters_changed(self):
		
		self.update()
	
	@QtCore.Slot()
	def on_store_attributes(self):
		
		obj_id = self.cmain.cgraph.get_selected_samples()
		if len(obj_id) != 1:
			return
		obj_id = obj_id[0]
		self.cmain.cmodel.store_attributes(obj_id)
		self.update_attribute_data()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def clear(self):
		
		self._view.descriptors.clear()
	
	def update(self):
		
		n_clusters, _ = self._view.clusters.get_limits()
		n_max = self.cmain.cmodel.get_n_samples()
		clustering_enabled = (n_max > 2)
		n_selected = len(self.cmain.cgraph.get_selected_samples())
		
		if clustering_enabled:
			values = [""] + list(range(2, n_max))
			self._view.clusters.set_n_clusters(values, n_clusters)
		else:
			self._view.clusters.set_n_clusters([])
		
		self._view.clusters.set_n_samples(n_max)
		
		self._view.clusters.set_n_found(self.cmain.cgraph.get_n_clusters())
		
		self._view.descriptors.set_load_drawings_enabled(True)
		self._view.distances.set_calculate_distances_enabled(self.cmain.cgraph.has_drawings())
		self._view.distances.set_delete_distances_enabled(self.cmain.cmodel.has_distance())
		self._view.clusters.set_clustering_enabled(clustering_enabled)
		self._view.clusters.set_limit_enabled(clustering_enabled & (n_clusters is None))
		self._view.clusters.set_update_tree_enabled(self.cmain.cgraph.has_clusters())
		self._view.clusters.set_add_cluster_enabled(n_selected > 0)
		self._view.clusters.set_rename_cluster_enabled(len(self.cmain.cgraph.get_selected_clusters()) == 1)
		self._view.clusters.set_delete_enabled(self.cmain.cgraph.has_clusters())
	
	def set_db_name(self, name):
		
		self._view.set_db_name(name)
	
	def set_folder(self, path, url = None):
		
		self._view.set_folder(path, url)
	
	def set_attributes(self, rows):
		# rows = [(label, ctrl_type, name), ...]
		
		self._view.attributes.populate(rows)
	
	def set_attribute_data(self, data):
		# return data = {name: value, ...}
		
		# data = {name: (value, items), ...}; items = [value, ...]
		values = self.cmain.cmodel.get_descriptor_values()
		attr_data = {}
		for name in data:
			attr_data[name] = (data[name], [])
		for name in values:
			attr_data[name] = (attr_data.get(name, (None, []))[0], values[name])
		
		self._view.attributes.set_data(attr_data)
	
	def get_attribute_data(self):
		# return data = {name: value, ...}
		
		return self._view.attributes.get_data()
	
	def update_attribute_data(self):
		
		self.set_attribute_data(self.get_attribute_data())




from ceramatch.controller.cgraph_nodes.node import Node

from ceramatch.utils.fnc_drawing import (render_drawing, get_profile_data)

from deposit.utils.fnc_serialize import (try_numeric)
from deposit import DGeometry

from PySide2 import (QtWidgets, QtCore, QtGui, QtSvg)
import numbers

class SampleNode(Node):
	
	def __init__(self, cgraph, node_id, label, drawing_data, picture = None):
		
		Node.__init__(self, cgraph, node_id, label)
		
		self._drawing_data = drawing_data
		
		if picture is not None:
			self._picture = picture
		else:
			self._picture = QtGui.QPicture()
			painter = QtGui.QPainter(self._picture)
			render_drawing(
				self._drawing_data, painter, self.cgraph.LINE_WIDTH, 
				scale = self.cgraph.SCALE_DRAWINGS, color = QtCore.Qt.black,
			)
			painter.end()
		
		self._selection_polygon = QtGui.QPolygonF(
			QtCore.QRectF(self._picture.boundingRect().marginsAdded(
				QtCore.QMargins(3, 3, 3, 3)
			))
		)
		self._selection_shape = QtGui.QPainterPath()
		self._selection_shape.addPolygon(self._selection_polygon)
	
	def copy(self):
		
		return SampleNode(
			self.cgraph, self.node_id, self.label, 
			self._drawing_data, self._picture,
		)
	
	def get_drawing_data(self):
		
		return self._drawing_data.copy()
	
	def set_drawing_data(self, data):
		
		self._drawing_data = data
	
	def get_profile_data(self):
		# returns (coords, radius)
		
		coords, radius = get_profile_data(self._drawing_data)
		
		return (coords, radius)
	
	def boundingRect(self):
		
		scale = self.cgraph.get_scale_factor()
		if scale < self.cgraph.SCALE_CUTOFF:
			return QtCore.QRectF()
		
		return self._selection_polygon.boundingRect()
	
	def center(self):
		
		return self.boundingRect().center()
	
	def shape(self):
		
		return self._selection_shape
	
	def paint(self, painter, option, widget):
		
		scale = self.cgraph.get_scale_factor()
		if scale < self.cgraph.SCALE_CUTOFF:
			return
		
		selected = False
		color = QtCore.Qt.black
		if option.state & QtWidgets.QStyle.State_Selected:
			color = QtCore.Qt.red
			selected = True
		
		rect = self.boundingRect()
		bgcolor = QtGui.QColor("white")
		bgcolor.setAlphaF(0.8)
		painter.setBrush(QtGui.QBrush(bgcolor))
		painter.setPen(QtGui.QPen(bgcolor, 1))
		painter.drawRect(rect)
		
		painter.setBrush(QtGui.QBrush(color))
		painter.setPen(QtGui.QPen(color, 1))
		painter.drawPicture(0, 0, self._picture)
		
		if selected:
			painter.setBrush(QtGui.QBrush())
			painter.drawRect(rect)
	
	def update_tooltip(self):
		
		buffer = QtCore.QBuffer()
		buffer.open(QtCore.QIODevice.WriteOnly)
		scale = self.cgraph.SCALE_TOOLTIP / self.cgraph.SCALE_DRAWINGS
		gen = QtSvg.QSvgGenerator()
		gen.setOutputDevice(buffer)
		painter = QtGui.QPainter(gen)
		painter.scale(scale, scale)
		rect = self.boundingRect().marginsAdded(QtCore.QMargins(10, 10, 10, 10))
		rect = QtCore.QRectF(rect.x(), rect.y(), rect.width()*0.8, rect.height())
		painter.setPen(QtGui.QPen(QtCore.Qt.white, 0))
		painter.setBrush(QtGui.QBrush(QtCore.Qt.white))
		painter.drawRect(rect)
		painter.setPen(QtGui.QPen(QtCore.Qt.black, 0.5))
		painter.drawPicture(0, 0, self._picture)
		painter.drawText(rect, QtCore.Qt.AlignCenter | QtCore.Qt.AlignBottom, self.label)
		painter.end()
		self._tool_tip = "<img src=\"data:image/png;base64,%s\">" % (bytes(buffer.data().toBase64()).decode())
	
	def on_hover(self, state):
		# state: True = Enter, False = Leave
		
		if self._tool_tip is None:
			self.update_tooltip()
		
		self.cgraph.on_hover(self, state)




from openpyxl import (load_workbook, Workbook)
from openpyxl.styles import (Font)
from collections import defaultdict
import csv
import os

def save_xlsx(header, rows, path):
	# header = [column, ...]
	# rows = [[value, ...], ...]
	
	wb = Workbook()
	wb.guess_types = True
	ws = wb.active
	ws.append(header)
	for i in range(len(header)):
		ws.cell(row = 1, column = i + 1).font = Font(bold = True)
	r = 2
	for row in rows:
		ws.append(row)
		for i in range(len(row)):
			ws.cell(row = r, column = i + 1).number_format = "@"
		r += 1
	wb.save(path)

def save_csv(header, rows, path):
	# header = [column, ...]
	# rows = [[value, ...], ...]
	
	with open(path, "w", newline = "") as f:
		writer = csv.writer(f, dialect=csv.excel, quoting=csv.QUOTE_ALL)
		writer.writerow(header)
		for row in rows:
			writer.writerow([str(val) for val in row])

def import_clusters_xlsx(path, sample_column, cluster_column):
	
	wb = load_workbook(filename = path, read_only = True)
	ws = None
	for sheet in wb.sheetnames:
		ws = wb[sheet]
		break
	if ws is None:
		return {}
	columns = {}
	for i, cell in enumerate(list(ws.iter_rows(max_row = 1))[0]):
		value = cell.value
		if value is not None:
			value = str(value).strip()
			if value:
				columns[value] = i
	if (sample_column not in columns) or (cluster_column not in columns):
		return {}
	
	clusters = defaultdict(set)
	for row in ws.iter_rows(min_row = 2):
		sample_id = str(row[columns[sample_column]].value).strip()
		cluster = str(row[columns[cluster_column]].value).strip()
		if sample_id and cluster:
			clusters[cluster].add(sample_id)
	
	return dict(clusters)

def import_clusters_csv(path, sample_column, cluster_column):
	
	columns = {}
	with open(path, "r", newline = "") as f:
		reader = csv.reader(f, dialect = csv.excel, quoting=csv.QUOTE_ALL)
		for row in reader:
			for i, value in enumerate(row):
				value = value.strip()
				if value:
					columns[value] = i
			break
	if (sample_column not in columns) or (cluster_column not in columns):
		return {}
	
	clusters = defaultdict(set)
	with open(path, "r", newline = "") as f:
		reader = csv.reader(f, dialect = csv.excel, quoting=csv.QUOTE_ALL)
		next(reader)
		for row in reader:
			sample_id = str(row[columns[sample_column]]).strip()
			cluster = str(row[columns[cluster_column]]).strip()
			if sample_id and cluster:
				clusters[cluster].add(sample_id)
	
	return dict(clusters)




from deposit_gui.dgui.dgraph_view import AbstractNode

from PySide2 import (QtWidgets, QtCore, QtGui)

class Node(AbstractNode):
	
	def __init__(
		self, cgraph, node_id, label = "", 
		radius = 10, font_family = "Calibri", font_size = 16,
	):
		
		AbstractNode.__init__(self, node_id, label, font_family, font_size)
		
		self.cgraph = cgraph
		self.radius = radius
		self._tool_tip = None
		self._moved = False
	
	def boundingRect(self):
		
		adjust = 2		
		return QtCore.QRectF(
			-self.radius - adjust, 
			-self.radius - adjust, 
			(2*self.radius) + 3 + adjust, 
			(2*self.radius) + 3 + adjust,
		)
	
	def center(self):
		
		return QtCore.QPointF(0, 0)
	
	def shape(self):
		
		path = QtGui.QPainterPath()
		path.addEllipse(-self.radius, -self.radius, 2*self.radius, 2*self.radius)
		return path
	
	def get_tool_tip(self):
		
		return self._tool_tip
	
	def paint(self, painter, option, widget):
		
		pen_width = 0
		if option.state & QtWidgets.QStyle.State_Sunken:
			pen_width = 2
		
		
		color = QtCore.Qt.white
		if (option.state & QtWidgets.QStyle.State_Selected):
			color = QtCore.Qt.red
		elif (option.state & QtWidgets.QStyle.State_HasFocus):
			color = QtCore.Qt.darkGray
		painter.setBrush(QtGui.QBrush(color))
		
		painter.setPen(QtGui.QPen(QtCore.Qt.black, pen_width))
		painter.drawEllipse(
			-self.radius, -self.radius, 2*self.radius, 2*self.radius
		)
	
	def on_hover(self, state):
		# state: True = Enter, False = Leave
		
		self.cgraph.on_hover(self, state)
	
	def on_position_change(self):
		
		self._moved = True
		for node in self.collidingItems():
			if not isinstance(node, Node):
				continue
			node.setFocus()
	
	def on_mouse_press(self):
		# re-implement
		
		pass
	
	def on_mouse_release(self):
		
		if self._moved:
			self._moved = False
			self.cgraph.on_moved(self)
		
		self.cgraph.on_mouse_released(self)
		
		for node in self.collidingItems():
			if not isinstance(node, Node):
				continue
			node.on_drop(self)
	
	def on_drop(self, node):
		
		self.cgraph.on_drop(node, self)




from ceramatch.controller.cmodel import CModel
from ceramatch.controller.cview import CView
from ceramatch.controller.cgraph import CGraph
from ceramatch.controller.ccontrols import CControls
from ceramatch.controller.cdialogs import CDialogs
from ceramatch.controller.cactions import CActions
from ceramatch.controller.chistory import CHistory

from ceramatch import __version__

from deposit_gui.controller.controller import Controller as DepositController

from PySide2 import (QtWidgets, QtCore, QtGui)
import json
import sys
import os

class Controller(QtCore.QObject):
	
	def __init__(self):
		
		QtCore.QObject.__init__(self)
		
		self._deposit_controller = None
		
		self.history = CHistory(self)
		self.cmodel = CModel(self)
		self.ccontrols = CControls(self)
		self.cgraph = CGraph(self)
		self.cview = CView(self, self.ccontrols, self.cgraph)
		self.cdialogs = CDialogs(self, self.cview)
		self.cactions = CActions(self, self.cview)
		
		self.cmodel.set_progress(self.cview.progress)
		self.cmodel.load_descriptors()
		
		self.cview.log_message("CeraMatch started")
		
		self.cview.show()
		
		self.cview.init_query_toolbar()
		
		self.ccontrols.update()
		self.cmodel.update_model_info()
		self.cdialogs.open("Connect")
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	def on_close(self):
		
		if not self.check_save():
			return False
		
		if self._deposit_controller is not None:
			self._deposit_controller.close()
		return True
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def clear(self):
		
		self.ccontrols.clear()
		self.cgraph.clear()
		self.history.clear()
	
	def check_save(self):
		
		if not self.cmodel.is_saved():
			reply = QtWidgets.QMessageBox.question(self.cview._view, 
				"Exit", "Save changes to database?",
				QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel
			)
			if reply == QtWidgets.QMessageBox.Yes:
				self.cactions.on_Save(True)
			elif reply == QtWidgets.QMessageBox.No:
				return True
			else:
				return False
		
		return True
	
	def import_descriptors(self, path):
		
		with open(path, "r") as f:
			data = json.load(f)
		if len(data) == 2:
			attributes, descriptors = data
			on_drawing = []
		else:
			attributes, descriptors, _ = data
		self.cmodel.set_descriptors(descriptors)
		self.cmodel.set_attributes(attributes)
		self.ccontrols.update_attribute_data()
	
	def open_deposit(self, querystr = None):
		
		if self._deposit_controller is None:
			self._deposit_controller = DepositController(
				self.cmodel._model._store
			)
		self._deposit_controller.cview.show()
		if querystr is not None:
			self._deposit_controller.cmdiarea.add_query(querystr)
	
	def open_folder(self, path):
		
		if sys.platform in ["linux", "linux2", "darwin"]:
			return # TODO
		if sys.platform.startswith("win"):
			if os.path.isdir(path):
				os.startfile(path)
	
	def open_in_external(self, path):
		
		if sys.platform in ["linux", "linux2", "darwin"]:
			return # TODO		
		if sys.platform.startswith("win"):
			if os.path.isfile(path):
				os.startfile(path)




from PySide2 import (QtWidgets, QtCore, QtGui)
import weakref

TOOLTIP_SHOW_DELAY = 800
TOOLTIP_HIDE_DELAY = 400

class ToolTip(QtWidgets.QLabel):
	
	def __init__(self):
		
		self._node = None
		
		QtWidgets.QLabel.__init__(self)
		
		self.setWindowFlags(QtCore.Qt.ToolTip)
		self.setFrameShape(QtWidgets.QFrame.StyledPanel)
		self.setStyleSheet("QLabel { background-color : white; }")
		
		self._hide_timer = QtCore.QTimer(singleShot = True, timeout = self.on_hide_timer)
		self._show_timer = QtCore.QTimer(singleShot = True, timeout = self.on_show_timer)
		
		self.hide()
	
	def move(self, pos):
		
		self.ensurePolished()
		geo = QtCore.QRect(pos, self.sizeHint())
		try:
			screen = QtWidgets.QApplication.screenAt(pos)
		except:
			for screen in QtWidgets.QApplication.screens():
				if pos in screen.geometry():
					break
			else:
				screen = None
		if not screen:
			screen = QtWidgets.QApplication.primaryScreen()
		screenGeo = screen.availableGeometry()
		if geo.bottom() > screenGeo.bottom():
			geo.moveBottom(screenGeo.bottom())
		if geo.top() < screenGeo.top():
			geo.moveTop(screenGeo.top())
		if geo.right() > screenGeo.right():
			geo.moveRight(screenGeo.right())
		if geo.left() < screenGeo.left():
			geo.moveLeft(screenGeo.left())
		QtWidgets.QLabel.move(self, geo.topLeft())
	
	def show(self, text, node):
		
		self._show_timer.stop()
		self._hide_timer.stop()
		self._node = weakref.ref(node)
		self.setText(text)
		self.adjustSize()
		self.move(QtGui.QCursor.pos())
		self._show_timer.start(TOOLTIP_SHOW_DELAY)
	
	def hide(self, delay = None):
		
		self._show_timer.stop()
		if delay is None:
			self._hide_timer.stop()
			QtWidgets.QLabel.hide(self)
			return
		if self.underMouse():
			return
		self._hide_timer.start(delay)
	
	@QtCore.Slot()
	def on_show_timer(self):
		
		node = self._node()
		if node is None:
			return
		if QtWidgets.QApplication.mouseButtons() != QtCore.Qt.NoButton:
			return
		if node.isUnderMouse():
			QtWidgets.QLabel.show(self)
	
	@QtCore.Slot()
	def on_hide_timer(self):
		
		self.hide()
	
	def leaveEvent(self, event):
		
		if self.isVisible():
			self.hide(TOOLTIP_HIDE_DELAY)
		else:
			self._show_timer.stop()
		
		QtWidgets.QLabel.leaveEvent(self, event)
	
	def mousePressEvent(self, event):
		
		self.hide()
		
		QtWidgets.QLabel.mousePressEvent(self, event)



from lap_data import LCModel

from deposit_gui import DRegistry

from deposit.datasource import (AbstractDatasource, Memory)
from deposit import (DDateTime, DGeometry, DResource)
from deposit.utils.fnc_files import (as_url)
from deposit.query.parse import (remove_bracketed_all)

from PySide2 import (QtCore)
from collections import defaultdict
from itertools import combinations
import datetime
import winreg
import json
import os


class CModel(LCModel):
	
	def __init__(self, cmain):
		
		LCModel.__init__(self, cmain)
		
		self._primary_class = None
		self._cm_classes = {
			"cluster": "CMCluster", 
			"node": "CMNode",
			"position": "CMPosition",
		}
		self._dist_rels = [
			"diam_dist", "axis_dist", "dice_dist", "dice_rim_dist"
		]
		self._distance = {}
		self._query = self.get_query("")
		self._objects = None
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	def on_set_descriptors(self, data):
		# data = [[name, chain], ...];
		# 	chain = "Class.Descriptor" or "Class.Relation.Class.Descriptor"
		
		pass
		
	def on_set_attributes(self, data):
		# data = [(label, ctrl_type, name), ...]
		
		self.cmain.ccontrols.set_attributes(data)
	
	def on_added(self, objects, classes):
		# elements = [DObject, DClass, ...]
		
		self.cmain.cactions.update()
		self.cmain.ccontrols.update()
	
	def on_deleted(self, objects, classes):
		# elements = [obj_id, name, ...]
		
		self.cmain.cactions.update()
		self.cmain.ccontrols.update()
	
	def on_changed(self, objects, classes):
		# elements = [DObject, DClass, ...]
		
		self.cmain.cactions.update()
		self.cmain.ccontrols.update()
	
	def on_saved(self, datasource):
		
		self.cmain.cview.set_status_message("Saved: %s" % (str(datasource)))
		self.cmain.cactions.update()
	
	def on_loaded(self):
		
		self.cmain.clear()
		self.update_model_info()
		self.cmain.ccontrols.on_loaded()
		self.cmain.cview.on_loaded()
		self.cmain.cactions.update()
	
	def on_settings_changed(self):
		
		self.cmain.cactions.update()
		self.cmain.ccontrols.update()
	
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def clear(self):
		
		self._distance = {}
		self._query = self.get_query("")
		self._objects = None
	
	def set_cm_classes(self, data):
		# data = {"cluster": name, "node": name, "position": name}
		
		self._cm_classes = data
	
	def load_descriptors(self):
		
		def _get_data(lap_registry, name):
			
			data = lap_registry.get(name)
			if not data:
				return []
			return json.loads(data)
		
		lap_4_found = False
		lap_3_found = False
		try:
			winreg.OpenKey(winreg.HKEY_CURRENT_USER, "SOFTWARE\LAP 4")
			lap_4_found = True
		except:
			lap_4_found = False
		if not lap_4_found:
			try:
				winreg.OpenKey(winreg.HKEY_CURRENT_USER, "SOFTWARE\Laser Aided Profiler")
				lap_3_found = True
			except:
				lap_3_found = False
		if lap_4_found or not lap_3_found:
			lap_registry = DRegistry("LAP 4")
		else:
			lap_registry = DRegistry("Laser Aided Profiler")
		
		self.set_descriptors(_get_data(lap_registry, "descriptors"))
		self.set_attributes(_get_data(lap_registry, "attributes"))
		self.set_cm_classes(
			self.cmain.cview.get_registry("cm_classes", self._cm_classes)
		)
	
	def get_primary_class(self):
		# returns class of Sample
		
		if self._primary_class is None:
			for name, chain in self._descriptors:
				if name == self.NAME_ID:
					self._primary_class = self.parse_chain(chain)[0]
					break
		
		return self._primary_class
	
	def get_cm_classes(self):
		# returns {"cluster": name, "node": name, "position": name}
		
		return self._cm_classes
	
	def set_query(self, querystr):
		
		self._distance = {}
		self._objects = None
		self._query.querystr = querystr
		self._query.process()
	
	def get_drawing_objects(self):
		
		if self._objects is None:
			self._objects = []
			for row in self._query:
				obj = self.get_object(row[0][0])
				if obj is not None:
					self._objects.append(obj)
		return self._objects
	
	def load_drawings(self):
		
		def _get_position(obj, descr_pos):
			
			pos = obj.get_descriptor(descr_pos)
			if isinstance(pos, DGeometry):
				x, y = pos.coords
				return (x, y)
			return None
		
		name_lookup, class_descriptors, primary_class, rel_primary, rel_chains = \
			self.get_load_lookups()
		
		for cls_del in ["Arc", "Photo", "Measure", "Annotation"]:
			if cls_del in class_descriptors:
				del class_descriptors[cls_del]
		
		recons_descr = None
		for cls, descr in name_lookup:
			if name_lookup[(cls, descr)] == "Reconstruction":
				recons_descr = descr
				break
		
		drawing_data = {}  
		# {obj_id: {key: value, key: [{key: value, ...}, ...], ...}, ...}
		cluster_data = {}
		# {obj_id: {"name": value, "position": (x, y), "children": [obj_id, ...], ...}, ...}
		node_data = {}
		# {obj_id: {"name": value, "position": (x, y), ...}, ...}
		edges = set()
		# [(source_id, target_id), ...]
		
		descr_pos = self._cm_classes["position"]
		objects = self.get_drawing_objects()
		if not objects:
			return
		
		self._progress.show("Loading Drawings")
		self._progress.update_state(value = 0, maximum = len(objects))
		cnt = 1
		for obj in objects:
			self._progress.update_state(value = cnt)
			cnt += 1
			if self._progress.cancel_pressed():
				break
			
			drawing_data[obj.id] = self.load_object_data(
				obj.id, name_lookup, class_descriptors, 
				primary_class, rel_primary, rel_chains,
			)
			pos = _get_position(obj, descr_pos)
			if pos is not None:
				drawing_data[obj.id]["position"] = pos
		
		self._distance = {}
		self._progress.show("Loading Distances")
		self._progress.update_state(value = 0, maximum = len(objects))
		cnt = 1
		done = set()
		for obj in objects:
			self._progress.update_state(value = cnt)
			cnt += 1
			if self._progress.cancel_pressed():
				break
			
			for obj_tgt, label in obj.get_relations():
				if label not in self._dist_rels:
					continue
				w = obj.get_relation_weight(obj_tgt.id, label)
				if w is None:
					continue
				idx = self._dist_rels.index(label)
				if obj.id not in self._distance:
					self._distance[obj.id] = {}
				if obj_tgt.id not in self._distance[obj.id]:
					self._distance[obj.id][obj_tgt.id] = [None, None, None, None]
				self._distance[obj.id][obj_tgt.id][idx] = w
				done.add((obj.id, obj_tgt.id))
		
		for obj_id1, obj_id2 in combinations([obj.id for obj in objects], 2):
			if ((obj_id1, obj_id2) in done) or ((obj_id2, obj_id1) in done):
				continue
			if obj_id1 not in self._distance:
				self._distance[obj_id1] = {}
			self._distance[obj_id1][obj_id2] = [None, None, None, None]
		
		cls_cluster = self.get_class(self._cm_classes["cluster"])
		cls_node = self.get_class(self._cm_classes["node"])
		cm_names = set([
			primary_class, self._cm_classes["cluster"], self._cm_classes["node"]
		])
		
		if cls_cluster is not None:
			for obj in cls_cluster.get_members(direct_only = True):
				pos = _get_position(obj, descr_pos)
				name = obj.get_descriptor("Name")
				if (pos is None) or (name is None):
					continue
				children = set()
				for obj2, label in obj.get_relations():
					if label != "contains":
						continue
					classes = obj2.get_class_names()
					if (primary_class in classes) and (obj2 not in objects):
						continue
					if cm_names.intersection(classes):
						children.add(obj2.id)
						edges.add((obj.id, obj2.id))
				cluster_data[obj.id] = dict(
					name = name,
					position = pos,
					children = children,
				)
		
		if cls_node is not None:
			for obj in cls_node.get_members(direct_only = True):
				if obj.id in cluster_data:
					continue
				pos = _get_position(obj, descr_pos)
				name = obj.get_descriptor("Name")
				if (pos is None) or (name is None):
					continue
				for obj2, label in obj.get_relations():
					if label != "linked":
						continue
					classes = obj2.get_class_names()
					if (primary_class in classes) and (obj2 not in objects):
						continue
					if cm_names.intersection(classes):
						edges.add((obj.id, obj2.id))
				node_data[obj.id] = dict(
					name = name,
					position = pos,
				)
		
		self._progress.stop()
		
		self.cmain.cgraph.populate(drawing_data, cluster_data, node_data, edges)
	
	def get_n_samples(self):
		
		return len(self._distance)
	
	def has_distance(self):
		
		for obj_id1 in self._distance:
			for obj_id2 in self._distance[obj_id1]:
				for val in self._distance[obj_id1][obj_id2]:
					if val is not None:
						return True
		return False
	
	def save_distance(self):
		
		self._model.blockSignals(True)
		changed_ids = set()
		for obj_id1 in self._distance:
			obj1 = self.get_object(obj_id1)
			if obj1 is None:
				continue
			changed_ids.add(obj_id1)
			changed_ids.update(self._distance[obj_id1].keys())
			for obj_id2 in self._distance[obj_id1]:
				dists = self._distance[obj_id1][obj_id2]
				for idx, label in enumerate(self._dist_rels):
					if dists[idx] is not None:
						obj1.add_relation(obj_id2, label, weight = dists[idx])
		self._model.blockSignals(False)
		if changed_ids:
			self._model.on_changed(changed_ids)
	
	def delete_distance(self):
		
		self._distance.clear()
		
		objects = self.get_drawing_objects()
		if not objects:
			return
		self._model.blockSignals(True)
		changed_ids = set()
		for obj in objects:
			to_del = set()
			for obj_tgt, label in obj.get_relations():
				if label not in self._dist_rels:
					continue
				to_del.add((obj_tgt.id, label))
				changed_ids.add(obj.id)
				changed_ids.add(obj_tgt.id)
			for obj_tgt_id, label in to_del:
				obj.del_relation(obj_tgt_id, label)
		self._model.blockSignals(False)
		self._model.on_changed(changed_ids)
	
	def add_cluster(self, label, children):
		
		cls_cluster = self.add_class(self._cm_classes["cluster"])
		obj_cluster = cls_cluster.add_member()
		obj_cluster.set_descriptor("Name", label)
		for obj_id in children:
			obj = self.get_object(obj_id)
			if obj is None:
				continue
			obj_cluster.add_relation(obj_id, "contains")
		
		return obj_cluster.id
	
	def add_node(self, label):
		
		cls_node = self.add_class(self._cm_classes["node"])
		obj_node = cls_node.add_member()
		obj_node.set_descriptor("Name", label)
		
		return obj_node.id
	
	def add_link(self, source_id, target_id):
		
		obj_src = self.get_object(source_id)
		obj_tgt = self.get_object(target_id)
		if (obj_src is None) or (obj_tgt is None):
			return
		obj_src.add_relation(target_id, "linked")
	
	def del_link(self, source_id, target_id):
		
		obj_src = self.get_object(source_id)
		obj_tgt = self.get_object(target_id)
		if (obj_src is None) or (obj_tgt is None):
			return
		to_del = set()
		for obj_tgt_, label in obj_src.get_relations():
			if label not in ["linked", "contains"]:
				continue
			if obj_tgt_ != obj_tgt:
				continue
			to_del.add(label)
		for label in to_del:
			obj_src.del_relation(obj_tgt.id, label)
	
	def set_cluster(self, sample_id, cluster_id):
		
		obj_sample = self.get_object(sample_id)
		obj_cluster = self.get_object(cluster_id)
		if (obj_sample is None) or (obj_cluster is None):
			return
		
		self._model.blockSignals(True)
		changed_ids = set([obj_sample.id, obj_cluster.id])
		to_del = set()
		for obj_tgt, label in obj_sample.get_relations():
			if self._cm_classes["cluster"] not in obj_tgt.get_class_names():
				continue
			to_del.add((obj_tgt.id, label))
		for obj_tgt_id, label in to_del:
			obj_sample.del_relation(obj_tgt_id, label)
			changed_ids.add(obj_tgt_id)
		
		obj_cluster.add_relation(sample_id, "contains")
		
		self._model.blockSignals(False)
		self._model.on_changed(changed_ids)
	
	def set_cluster_label(self, obj_id, label):
		
		obj = self.get_object(obj_id)
		if obj is None:
			return
		obj.set_descriptor("Name", label)
	
	def delete_clusters(self, object_ids = None):
		
		all_objects = self.get_drawing_objects()
		
		if object_ids is None:
			objects = all_objects
		else:
			objects = []
			for obj_id in object_ids:
				obj = self.get_object(obj_id)
				if obj is not None:
					objects.append(obj)
		if not objects:
			return
		
		self._model.blockSignals(True)
		deleted_ids = set()
		changed_ids = set()
		
		for obj in objects:
			to_del = set()
			for obj_tgt, label in obj.get_relations():
				if self._cm_classes["cluster"] not in obj_tgt.get_class_names():
					continue
				to_del.add((obj_tgt.id, label))
			for obj_tgt_id, label in to_del:
				obj.del_relation(obj_tgt_id, label)
				changed_ids.add(obj.id)
				changed_ids.add(obj_tgt_id)
		
		cls_cluster = self.get_class(self._cm_classes["cluster"])
		cls_node = self.get_class(self._cm_classes["node"])
		
		if cls_cluster is not None:
			primary_class = self.get_primary_class()
			to_del = set()
			for obj in cls_cluster.get_members(direct_only = True):
				found_rel = False
				for obj_tgt, label in obj.get_relations():
					if primary_class not in obj_tgt.get_class_names():
						continue
					found_rel = True
					break
				if not found_rel:
					to_del.add(obj.id)
			if to_del:
				self.del_objects(to_del)
				deleted_ids.update(to_del)
		
		if cls_node is not None:
			while True:
				to_del = set()
				for obj in cls_node.get_members(direct_only = True):
					found_rel = False
					for obj_tgt, label in obj.get_relations():
						if label in ["linked", "contains"]:
							found_rel = True
							break
					if not found_rel:
						to_del.add(obj.id)
				if to_del:
					self.del_objects(to_del)
					deleted_ids.update(to_del)
				else:
					break
		
		changed_ids = changed_ids.difference(deleted_ids)
		
		self._model.blockSignals(False)
		if deleted_ids:
			self._model.on_deleted(deleted_ids)
		if changed_ids:
			self._model.on_changed(changed_ids)
		
		labels = {}  # {node_id: label, ...}
		objects = []
		_, sample_descr = self.get_cls_descr(self.NAME_ID)
		node_idxs = list(range(len(all_objects)))
		for node_idx, obj in enumerate(all_objects):
			objects.append(obj.id)
			labels[node_idx] = obj.get_descriptor(sample_descr)
		
		all_objects = [obj.id for obj in all_objects]
		self.cmain.cgraph.set_clusters(all_objects, {}, node_idxs, [], labels)
	
	def save_position(self, obj_id, x, y):
		
		descr_pos = self._cm_classes["position"]
		obj = self.get_object(obj_id)
		if obj is None:
			return
		obj.set_geometry_descriptor(self._cm_classes["position"], "POINT", [x, y])
	
	def store_attributes(self, obj_id):
		
		descr_lookup = dict(self.get_descriptors())
		attribute_data = self.cmain.ccontrols.get_attribute_data()
		data = {}
		for name in attribute_data:
			if name in descr_lookup:
				data[(name, descr_lookup[name])] = attribute_data[name]
		
		data[("Date_Modified", descr_lookup["Date_Modified"])] = DDateTime(
			datetime.datetime.now()
		)
		
		# data = {(name, chain): value, key: [{(name, chain): value, ...}, ...], ...}
		self.store_data(data, obj_id, keep_default = True)
		
		self.cmain.cgraph.update_sample_data(obj_id, attribute_data)
	
	
	# ---- Deposit
	# ------------------------------------------------------------------------
	def update_model_info(self):
		
		self.cmain.cview.set_title(self.get_datasource_name())
		self.cmain.ccontrols.set_db_name("%s (%s)" % (
			self.get_datasource_name(),
			str(self.get_datasource()),
		))
		
		path = self.get_folder()
		url = as_url(path)
		if isinstance(self.get_datasource(), Memory):
			path = "temporary"
		self.cmain.ccontrols.set_folder(path, url)
	
	def update_recent(self, kwargs):
		
		datasource = kwargs.get("datasource", None)
		if isinstance(datasource, AbstractDatasource):
			kwargs.update(datasource.to_dict())
		
		url = kwargs.get("url", None)
		if not url:
			path = kwargs.get("path", None)
			if path:
				url = as_url(path)
		self.cmain.cview._view.add_recent_connection(
			url = url,
			identifier = kwargs.get("identifier", None),
			connstr = kwargs.get("connstr", None),
		)
	
	def load(self, *args, **kwargs):
		# datasource = Datasource or format
		
		if not self.cmain.check_save():
			return False
		
		self.clear()
		
		return LCModel.load(self, *args, **kwargs)




# -*- coding: utf-8 -*-
#!/usr/bin/env python
#

from setuptools import setup, find_packages
import pathlib

try:
	from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
	class bdist_wheel(_bdist_wheel):
		def finalize_options(self):
			_bdist_wheel.finalize_options(self)
			self.root_is_pure = False
except ImportError:
	bdist_wheel = None

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
	name="ceramatch",
	version="1.0.41",
	description="Visual shape-matching and classification of ceramics.",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/demjanp/CeraMatch",
	author="Peter Demján",
	author_email="peter.demjan@gmail.com",
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Science/Research",
		"Topic :: Scientific/Engineering",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Programming Language :: Python :: 3.10",
		"Operating System :: Microsoft :: Windows :: Windows 10",
	],
	keywords="ceramics, classification, cluster analysis, morphometrics",
	package_dir={"": "src"},
	packages=find_packages(where="src"),
	include_package_data=True,
	package_data={"":[
		"LICENSE",
		"*.TXT",
		"*.pyd",
		"*.png",
		"*.svg",
	]},
	python_requires=">=3.10, <4",
	install_requires=[
		'lap_data>=4.0.1, <4.1',
		'scipy>=1.9.1, <2',
		'scikit-image',
		'scikit-learn>=1.1.1, <1.2',
		'opencv-python>=4.5.5, <5',
		'numpy',
	],
	cmdclass={'bdist_wheel': bdist_wheel},
)




from PySide2 import (QtWidgets, QtCore, QtGui)

class Combo(QtWidgets.QComboBox):
	
	def __init__(self, callback, editable = False):
		
		QtWidgets.QComboBox.__init__(self)
		
		self.currentTextChanged.connect(callback)
		self.setEditable(editable)
	
	def clear_values(self):
		
		self.blockSignals(True)
		self.clear()
		self.blockSignals(False)
	
	def set_values(self, values, default = None):
		
		values = [str(val) for val in values]
		if (default is not None) and (default not in values):
			values = [default] + values
		value = self.currentText()
		if value:
			default = value
		self.blockSignals(True)
		self.clear()
		self.addItems(values)
		if default in values:
			self.setCurrentIndex(values.index(default))
		self.blockSignals(False)
	
	def get_value(self):
		
		return self.currentText()



from ceramatch.view.vcontrols_groups.controls.button import (Button)

from PySide2 import (QtWidgets, QtCore, QtGui)
from deposit.utils.fnc_files import (as_url, url_to_path)

class AttributeControl(object):
	
	signal_changed = QtCore.Signal(str, str)  # (name, value)
	
	def __init__(self, name):
		
		self.name = name
	
	def get_value(self):
		
		return ""
	
	def set_value(self, value):
		
		pass
	
	def set_items(self, values):
		
		pass
	
	@QtCore.Slot()
	def on_changed(self, *args):
		
		self.signal_changed.emit(self.name, self.get_value())

class LineEdit(AttributeControl, QtWidgets.QLineEdit):
	
	def __init__(self, name):
		
		AttributeControl.__init__(self, name)
		QtWidgets.QLineEdit.__init__(self)
		
		self.textChanged.connect(self.on_changed)
	
	def get_value(self):
		
		return self.text().strip()
	
	def set_value(self, value):
		
		self.blockSignals(True)
		self.setText(value)
		self.blockSignals(False)

class ComboBox(AttributeControl, QtWidgets.QComboBox):
	
	def __init__(self, name):
		
		AttributeControl.__init__(self, name)
		QtWidgets.QComboBox.__init__(self)
		
		self.setEditable(True)
		
		self.currentTextChanged.connect(self.on_changed)
	
	def get_value(self):
		
		return self.currentText().strip()
	
	def set_value(self, value):
		
		self.blockSignals(True)
		self.setCurrentText(value)
		self.blockSignals(False)
	
	def set_items(self, values):
		
		self.blockSignals(True)
		current_value = self.currentText()
		self.clear()
		if values:
			self.addItems(values)
		if current_value in values:
			self.setCurrentIndex(values.index(current_value))
		elif current_value:
			self.setCurrentText(current_value)
		self.blockSignals(False)

class CheckBox(AttributeControl, QtWidgets.QCheckBox):
	
	def __init__(self, name):
		
		
		AttributeControl.__init__(self, name)
		QtWidgets.QCheckBox.__init__(self)
		
		self.stateChanged.connect(self.on_changed)
	
	def get_value(self):
		
		return str(int(self.isChecked()))
	
	def set_value(self, value):
		
		value_ = False
		try:
			value_ = bool(int(value))
		except:
			raise Exception(
				"Error: Could not convert attribute %s, value '%s' to bool" % (self.name, str(value))
			)
		self.blockSignals(True)
		self.setChecked(value_)
		self.blockSignals(False)
	
	def set_items(self, values):
		
		pass

CONTROL_CLASSES = {
	"LineEdit": LineEdit,
	"ComboBox": ComboBox,
	"CheckBox": CheckBox,
}

class Attributes(QtWidgets.QGroupBox):
	
	signal_store_attributes = QtCore.Signal()
	
	def __init__(self):
		
		QtWidgets.QGroupBox.__init__(self, "Attributes")
		
		self._attributes = {} #  {lap_name: control, ...}
		self._labels = {}  # {lap_name: label, ...}
		self._attributes_changed = False
		
		self.setStyleSheet("QGroupBox {font-weight: bold;}")
		self.setLayout(QtWidgets.QVBoxLayout())
		
		self._form_frame = QtWidgets.QFrame()
		self._form_frame.setLayout(QtWidgets.QFormLayout())
		self._form_frame.layout().setContentsMargins(5, 5, 5, 5)
		
		self.store_button = Button("Store", self.on_store)
		
		label = QtWidgets.QLabel(
			"To modify attributes, use the LAP application - 'Edit Descriptors' function."
		)
		label.setWordWrap(True)
		
		self.layout().addWidget(self._form_frame)
		self.layout().addWidget(self.store_button)
		self.layout().addWidget(label)
		
		self.update_store_button()
	
	
	# ---- Signal handling
	# ------------------------------------------------------------------------
	@QtCore.Slot(str, str)
	def on_changed(self, name, value):
		
		self._attributes_changed = True
		self.update_store_button()
	
	@QtCore.Slot(str, str)
	def on_store(self):
		
		self.signal_store_attributes.emit()
	
	# ---- get/set
	# ------------------------------------------------------------------------
	def update_store_button(self):
		
		state = False
		if self._attributes_changed and self.get_data():
			state = True
		self.store_button.setEnabled(state)
	
	def populate(self, rows):
		# rows = [(label, ctrl_type, name), ...]
		
		self._attributes_changed = False
		self._attributes = {}
		self._labels = {}
		layout = self._form_frame.layout()
		for row in reversed(range(layout.rowCount())):
			layout.removeRow(row)
		for label, ctrl_type, name in rows:
			if ctrl_type not in CONTROL_CLASSES:
				continue
			self._attributes[name] = CONTROL_CLASSES[ctrl_type](name)
			self._attributes[name].signal_changed.connect(self.on_changed)
			self._labels[name] = label
			layout.addRow(QtWidgets.QLabel("%s:" % label), self._attributes[name])
		self.update_store_button()
	
	def clear(self):
		
		self.populate([])
		self.update_store_button()
	
	def get_data(self):
		# return data = {name: value, ...}
		
		data = {}
		for name in self._attributes:
			value = self._attributes[name].get_value()
			if value:
				data[name] = value
		
		return data
	
	def set_data(self, data):
		# data = {name: (value, items), ...}; items = [value, ...]
		
		self._attributes_changed = False
		for name in self._attributes:
			value, items = data.get(name, ("", []))
			if value is None:
				value = ""
			self._attributes[name].set_items([str(value) for value in items])
			self._attributes[name].set_value(str(value))
		self.update_store_button()



from ceramatch.controller.cgraph_nodes.node import Node

from PySide2 import (QtWidgets, QtCore, QtGui, QtSvg)

class ClusterNode(Node):
	
	def __init__(self, cgraph, node_id, label, children):
		
		Node.__init__(self, cgraph, node_id, label)
		
		self._label_rect = QtGui.QFontMetrics(self.font).boundingRect(self.label)
		self._label_rect.moveLeft(-self._label_rect.center().x())
		self._label_rect.moveTop(self.radius)
	
	def set_label(self, label):
		
		self.label = str(label)
		self._label_rect = QtGui.QFontMetrics(self.font).boundingRect(self.label)
		self._label_rect.moveLeft(-self._label_rect.center().x())
		self._label_rect.moveTop(self.radius)
		self.update(self.boundingRect())
	
	def paint(self, painter, option, widget):
		
		scale = self.cgraph.get_scale_factor()
		color = QtCore.Qt.lightGray
		if (option.state & QtWidgets.QStyle.State_Selected):
			color = QtCore.Qt.red
		elif (option.state & QtWidgets.QStyle.State_HasFocus):
			color = QtCore.Qt.darkGray
		pen_width = 0
		if option.state & QtWidgets.QStyle.State_Sunken:
			pen_width = 2
		painter.setBrush(QtGui.QBrush(color))
		painter.setPen(QtGui.QPen(QtCore.Qt.darkGray, pen_width))
		radius = self.radius / scale
		painter.drawEllipse(-radius, -radius, 2*radius, 2*radius)
		if scale > 0.56:
			painter.setFont(self.font)
			painter.drawText(
				self._label_rect, 
				QtCore.Qt.AlignCenter | QtCore.Qt.AlignBottom, 
				self.label,
			)
	
	def boundingRect(self):
		
		scale = self.cgraph.get_scale_factor()
		radius = self.radius / scale
		return QtCore.QRectF(-radius, -radius, 2*radius, 2*radius).united(self._label_rect)
	
	def update_tooltip(self, gap = 5):
		
		nodes = list(self.get_children())
		n_side = len(nodes)
		if n_side > 3:
			n_side = max(1, int(n_side**0.5))
		x = 0
		y = 0
		h_max = 0
		n = 0
		w = 0
		x_max, y_max = 0, 0
		positions = []
		for node in nodes:
			positions.append([x + gap, y + gap])
			rect = node.boundingRect()
			w, h = rect.width() + gap, rect.height() + gap
			h_max = max(h_max, h)
			x += w
			n += 1
			if n > n_side:
				x = 0
				y += h_max
				h_max = 0
				n = 0
			x_max = max(x_max, x + w)
			y_max = max(y_max, y + h)
		if n_side == len(nodes):
			x_max -= w
		
		scale = self.cgraph.SCALE_TOOLTIP / self.cgraph.SCALE_DRAWINGS
		buffer = QtCore.QBuffer()
		buffer.open(QtCore.QIODevice.WriteOnly)
		gen = QtSvg.QSvgGenerator()
		gen.setOutputDevice(buffer)
		painter = QtGui.QPainter(gen)
		painter.scale(scale, scale)
		painter.setPen(QtGui.QPen(QtCore.Qt.white, 0))
		painter.setBrush(QtGui.QBrush(QtCore.Qt.white))
		rect = QtCore.QRectF(0, 0, x_max, y_max)
		painter.drawRect(rect)
		painter.setPen(QtGui.QPen(QtCore.Qt.black, 0.5))
		for i in range(len(nodes)):
			x, y = positions[i]
			painter.drawPicture(x, y, nodes[i]._picture)
		
		painter.drawText(rect, QtCore.Qt.AlignCenter | QtCore.Qt.AlignBottom, self.label)
		painter.end()
		self._tool_tip = "<img src=\"data:image/png;base64,%s\">" % (bytes(buffer.data().toBase64()).decode())
	
	def on_hover(self, state):
		# state: True = Enter, False = Leave
		
		if self._tool_tip is None:
			self.update_tooltip()
		
		self.cgraph.on_hover(self, state)
